/****************************************************************\
*                                                                *
*  C4 dynamic programming library - Seeded Dynamic Programming   *
*                                                                *
*  Guy St.C. Slater..   mailto:guy@ebi.ac.uk                     *
*  Copyright (C) 2000-2005.  All Rights Reserved.                *
*                                                                *
*  This source code is distributed under the terms of the        *
*  GNU Lesser General Public License. See the file COPYING       *
*  or http://www.fsf.org/copyleft/lesser.html for details        *
*                                                                *
*  If you use this code, please keep this notice intact.         *
*                                                                *
\****************************************************************/

#include <stdlib.h> /* For qsort() */

#include "sdp.h"
#include "matrix.h"

/**/

/* FIXME: temp */

static SDP_Row *SDP_Row_create(gint target_pos,
                               SDP_Scheduler *scheduler);
static void SDP_Scheduler_align_rows(SDP_Scheduler *scheduler);
static gboolean SDP_Scheduler_is_valid(SDP_Scheduler *scheduler);

/**/

static SDP_Cache_SlotData *SDP_Cache_SlotData_create(
                    SDP_Cache_FreeFunc free_func, gpointer free_data){
    register SDP_Cache_SlotData *slot_data =
        g_new(SDP_Cache_SlotData, 1);
    slot_data->free_func = free_func;
    slot_data->free_data = free_data;
    return slot_data;
    }

static void SDP_Cache_SlotData_destroy(SDP_Cache_SlotData *slot_data){
    g_free(slot_data);
    return;
    }

/**/

static void SDP_Cache_fill_func(SparseCache_Chunk *scc,
                                gpointer user_data){
    register SDP_Cache *sdp_cache = user_data;
    register gint i;
    for(i = 0; i < scc->length; i++)
        scc->data[i] = g_new0(gpointer, sdp_cache->slot_data_list->len);
    return;
    }

static void SDP_Cache_empty_func(SparseCache_Chunk *scc,
                                 gpointer user_data){
    register gint i, j;
    register SDP_Cache *sdp_cache = user_data;
    register gpointer *cache_data;
    register SDP_Cache_SlotData *slot_data;
    for(i = 0; i < scc->length; i++){
        cache_data = scc->data[i];
        for(j = 0; j < sdp_cache->slot_data_list->len; j++){
            g_assert(sdp_cache->slot_data_list->pdata[j]);
            slot_data = sdp_cache->slot_data_list->pdata[j];
            if(slot_data->free_func && cache_data[j]){
                slot_data->free_func(cache_data[j],
                                     slot_data->free_data);
                }
            }
        g_free(scc->data[i]);
        }
    return;
    }

static SDP_Cache *SDP_Cache_create(gint length){
    register SDP_Cache *sdp_cache = g_new(SDP_Cache, 1);
    sdp_cache->cache = SparseCache_create(length,
            SDP_Cache_fill_func, SDP_Cache_empty_func,
            sdp_cache, (1 << 12));
    sdp_cache->slot_data_list = g_ptr_array_new();
    return sdp_cache;
    }

static void SDP_Cache_destroy(SDP_Cache *sdp_cache){
    register gint i;
    register SDP_Cache_SlotData *slot_data;
    g_assert(sdp_cache);
    SparseCache_destroy(sdp_cache->cache);
    for(i = 0; i < sdp_cache->slot_data_list->len; i++){
        g_assert(sdp_cache->slot_data_list->pdata[i]);
        slot_data = sdp_cache->slot_data_list->pdata[i];
        SDP_Cache_SlotData_destroy(slot_data);
        }
    g_ptr_array_free(sdp_cache->slot_data_list, TRUE);
    g_free(sdp_cache);
    return;
    }

static gint SDP_Cache_add_slot(SDP_Cache *sdp_cache,
                               SDP_Cache_FreeFunc free_func,
                               gpointer free_data){
    register SDP_Cache_SlotData *slot_data
        = SDP_Cache_SlotData_create(free_func, free_data);
    g_ptr_array_add(sdp_cache->slot_data_list, slot_data);
    return sdp_cache->slot_data_list->len-1;
    }

static gpointer SDP_Cache_get(SDP_Cache *sdp_cache,
                              gint slot, gint pos){
    register gpointer *cache_data = SparseCache_get(sdp_cache->cache,
                                                    pos);
    return cache_data[slot];
    }

static void SDP_Cache_set(SDP_Cache *sdp_cache,
                          gint slot, gint pos, gpointer data){
    register gpointer *cache_data = SparseCache_get(sdp_cache->cache,
                                                    pos);
    cache_data[slot] = data;
    return;
    }

/**/

SDP *SDP_create(C4_Model *model){
    register SDP *sdp = g_new(SDP, 1);
    register C4_Portal *portal;
    /**/
    sdp->ref_count = 1;
    g_assert(model);
    g_assert(!model->is_open);
    g_assert(Lookahead_Mask_WIDTH > model->max_query_advance);
    g_assert(Lookahead_Mask_WIDTH > model->max_target_advance);
    sdp->model = C4_Model_share(model);
    /* Perform bidirectional SDP only
     * when there is a single match state and no shadows
     */
    sdp->use_boundary = TRUE;
    if(model->shadow_list->len == 0){
        if(model->portal_list->len == 1){
            portal = model->portal_list->pdata[0];
            g_assert(portal);
            if(portal->transition_list->len == 1)
                sdp->use_boundary = FALSE;
            }
        }
    sdp->pq_set = PQueueSet_create();
    return sdp;
    }

SDP *SDP_share(SDP *sdp){
    g_assert(sdp);
    sdp->ref_count++;
    return sdp;
    }

void SDP_destroy(SDP *sdp){
    g_assert(sdp);
    if(--sdp->ref_count)
        return;
    C4_Model_destroy(sdp->model);
    PQueueSet_destroy(sdp->pq_set);
    g_free(sdp);
    return;
    }

/**/

static gint SDP_ObstructList_compare(gconstpointer a,
                                     gconstpointer b){
    register SDP_Seed *seed_a = (SDP_Seed*)a, *seed_b = (SDP_Seed*)b;
    return seed_a->seed_id - seed_b->seed_id;
    }

static SDP_ObstructList *SDP_ObstructList_create(void){
    register SDP_ObstructList *obstruct_list
     = g_new(SDP_ObstructList, 1);
    obstruct_list->obstruct_tree = g_tree_new(SDP_ObstructList_compare);
    obstruct_list->last_added = NULL;
    return obstruct_list;
    }

static void SDP_ObstructList_destroy(SDP_ObstructList *obstruct_list){
    g_assert(obstruct_list);
    g_tree_destroy(obstruct_list->obstruct_tree);
    g_free(obstruct_list);
    return;
    }

static void SDP_ObstructList_add(SDP_ObstructList *obstruct_list,
                                 SDP_Seed *seed){
    if(obstruct_list->last_added == seed)
        return;
    obstruct_list->last_added = seed;
    g_tree_insert(obstruct_list->obstruct_tree, seed, seed);
    /* If key is already in the tree, no change is made */
    return;
    }

static gint SDP_ObstructList_traverse(gpointer key,
                                      gpointer value,
                                      gpointer data){
    register GPtrArray *list = data;
    g_ptr_array_add(list, value);
    return FALSE;
    }

static GPtrArray *SDP_ObstructList_get_list(
                  SDP_ObstructList *obstruct_list){
    register GPtrArray *list = g_ptr_array_new();
    g_tree_traverse(obstruct_list->obstruct_tree,
                    SDP_ObstructList_traverse, G_IN_ORDER, list);
    return list;
    }

/**/

static SDP_Terminal *SDP_Terminal_create(void){
    register SDP_Terminal *terminal = g_new(SDP_Terminal, 1);
    terminal->query_pos = 0;
    terminal->target_pos = 0;
    terminal->score = C4_IMPOSSIBLY_LOW_SCORE;
    terminal->obstruct_list = SDP_ObstructList_create();
    return terminal;
    }

static void SDP_Terminal_destroy(SDP_Terminal *terminal){
    SDP_ObstructList_destroy(terminal->obstruct_list);
    g_free(terminal);
    return;
    }

/**/

static SDP_Seed *SDP_Seed_create(HSP *hsp){
    register SDP_Seed *seed = g_new(SDP_Seed, 1);
    seed->seed_id = -1;
    seed->hsp = hsp;
    seed->max_start = SDP_Terminal_create();
    seed->max_end = SDP_Terminal_create();
    seed->pq_node = NULL;
    return seed;
    }
/* FIXME: optimisation: use RecycleBins for SDP_Terminal allocations
 */

static void SDP_Seed_destroy(SDP_Seed *seed){
    SDP_Terminal_destroy(seed->max_start);
    SDP_Terminal_destroy(seed->max_end);
    g_free(seed);
    return;
    }

/**/

static SDP_SpanSeed *SDP_SpanSeed_create(C4_Score score,
                         C4_Score max, gint seed_id,
                         gint query_start, gint target_start,
                         gint query_entry, gint target_entry,
                         C4_Score *shadow_data,
                         SDP_Scheduler *scheduler){
    register SDP_SpanSeed *span_seed = g_new(SDP_SpanSeed, 1);
    register gint i;
    register C4_Model *model = scheduler->sdp_pair->sdp->model;
    span_seed->score = score;
    span_seed->max = max;
    span_seed->seed_id = seed_id;
    span_seed->query_start = query_start;
    span_seed->target_start = target_start;
    span_seed->query_entry = query_entry;
    span_seed->target_entry = target_entry;
    if(scheduler->is_forward && model->total_shadow_designations){
        span_seed->shadow_data = g_new0(C4_Score,
                                 model->total_shadow_designations);
        if(shadow_data)
            for(i = 0; i < model->total_shadow_designations; i++)
                span_seed->shadow_data[i] = shadow_data[i];
    } else {
        span_seed->shadow_data = NULL;
        }
    return span_seed;
    }

static void SDP_SpanSeed_destroy(SDP_SpanSeed *span_seed){
    if(span_seed->shadow_data)
        g_free(span_seed->shadow_data);
    g_free(span_seed);
    return;
    }

static void SDP_SpanSeed_copy(SDP_SpanSeed *src, SDP_SpanSeed *dst,
                             SDP_Scheduler *scheduler){
    register gint i;
    register C4_Model *model = scheduler->sdp_pair->sdp->model;
    dst->score        = src->score;
    dst->max          = src->max;
    dst->seed_id      = src->seed_id;
    dst->query_start  = src->query_start;
    dst->target_start = src->target_start;
    dst->query_entry  = src->query_entry;
    dst->target_entry = src->target_entry;
    /* Copy shadows */
    if(scheduler->is_forward)
        for(i = 0; i < model->total_shadow_designations; i++)
            dst->shadow_data[i] = src->shadow_data[i];
    return;
    }

static SDP_SpanSeed *SDP_SpanSeed_duplicate(SDP_SpanSeed *span_seed,
                                            SDP_Scheduler *scheduler){
    return SDP_SpanSeed_create(span_seed->score,
                               span_seed->max,
                               span_seed->seed_id,
                               span_seed->query_start,
                               span_seed->target_start,
                               span_seed->query_entry,
                               span_seed->target_entry,
                               span_seed->shadow_data,
                               scheduler);
    }

/**/

static void SDP_SpanData_seed_free_func(gpointer data,
                                        gpointer user_data){
    register SDP_SpanSeed *span_seed = data;
    SDP_SpanSeed_destroy(span_seed);
    return;
    }

static SDP_SpanData *SDP_SpanData_create(SListSet *slist_set,
                                         C4_Span *span,
                                         SDP_Cache *cache){
    register SDP_SpanData *span_data = g_new(SDP_SpanData, 1);
    span_data->span = span;
    span_data->curr_span_seed = NULL;
    span_data->cache_slot = SDP_Cache_add_slot(cache,
                            SDP_SpanData_seed_free_func, NULL);
    return span_data;
    }

static void SDP_SpanData_destroy(SDP_SpanData *span_data){
    g_free(span_data);
    return;
    }

static SDP_SpanSeed *SDP_SpanData_get_curr(SDP_SpanData *span_data,
                                           SDP_Cache *sdp_cache,
                                           gint query_pos,
                                           gint target_pos){
    register SDP_SpanSeed *stored_seed;
    g_assert(span_data);
    g_assert(sdp_cache);
    /* Remove curr if expired */
    if(span_data->curr_span_seed){
        if((span_data->curr_span_seed->query_entry > query_pos)
        || ((span_data->curr_span_seed->query_entry
          + span_data->span->max_query) < query_pos)
        || ((span_data->curr_span_seed->target_entry
          + span_data->span->max_target) < target_pos)){
            span_data->curr_span_seed = NULL;
            }
        }
    /* Challenge curr with stored */
    stored_seed = SDP_Cache_get(sdp_cache,
                                span_data->cache_slot, query_pos);
    if(stored_seed){
       g_assert(stored_seed->query_entry == query_pos);
       if((stored_seed->target_entry + span_data->span->max_target)
        > target_pos){ /* Still valid */
            if(span_data->curr_span_seed){
                if(span_data->curr_span_seed->score
                  < stored_seed->score)
                    span_data->curr_span_seed = stored_seed;
            } else {
                span_data->curr_span_seed = stored_seed;
                }
        } else { /* Stored has expired */
            SDP_SpanSeed_destroy(stored_seed);
            SDP_Cache_set(sdp_cache, span_data->cache_slot,
                          query_pos, NULL);
            }
        }
    return span_data->curr_span_seed;
    }
/* SDP_SpanData_get_curr()
 * is called with transition from a span state
 * returns the best span_seed in scope
 */

/* Algorithm:
 * ---------
 * if(qy_span) challenging curr
 * if(ty_span) challenging cached
 * if(qy_span && ty_span)
 *    challenging curr and stored
 */

static void SDP_SpanData_submit(SDP_SpanData *span_data,
                                SDP_SpanSeed *span_seed,
                                SDP_Cache *sdp_cache,
                                SDP_Scheduler *scheduler){
    register SDP_SpanSeed *stored_seed;
    SDP_SpanData_get_curr(span_data, sdp_cache,
                          span_seed->query_entry,
                          span_seed->target_entry);
    if(span_data->span->max_query){ /* Challenge curr */
        if(span_data->curr_span_seed){
            if(span_data->curr_span_seed->score < span_seed->score){
                span_data->curr_span_seed
                    = SDP_SpanSeed_duplicate(span_seed, scheduler);
                }
        } else { /* No curr */
            SDP_SpanSeed_copy(span_seed, span_data->curr_span_seed,
                              scheduler);
            }
        }
    if(span_data->span->max_target){ /* Challenge cache */
        stored_seed = SDP_Cache_get(sdp_cache,
                                    span_data->cache_slot,
                                    span_seed->query_entry);
        if(stored_seed){
            g_assert(stored_seed->query_entry
                  == span_seed->query_entry);
            if(stored_seed->score < span_seed->score){
                SDP_SpanSeed_copy(span_seed, stored_seed, scheduler);
                }
        } else {
            SDP_Cache_set(sdp_cache,
                          span_data->cache_slot,
                          span_seed->query_entry,
                          SDP_SpanSeed_duplicate(span_seed, scheduler));
            }
        }
    return;
    }
/* SDP_SpanData_submit()
 * is called with transition to a span state
 */

/**/

static SDP_SpanHistory *SDP_SpanHistory_create(SListSet *slist_set,
                                               C4_Model *model,
                                               SDP_Cache *query_cache){
    register gint i;
    register SDP_SpanHistory *span_history = g_new(SDP_SpanHistory, 1);
    register SDP_SpanData *span_data;
    register C4_Span *span;
    span_history->model = C4_Model_share(model);
    span_history->span_data = g_new(SDP_SpanData*,
                                    model->span_list->len);
    for(i = 0; i < model->span_list->len; i++){
        span = model->span_list->pdata[i];
        span_data = SDP_SpanData_create(slist_set, span,
                                        query_cache);
        span_history->span_data[i] = span_data;
        }
    return span_history;
    }

static void SDP_SpanHistory_destroy(SDP_SpanHistory *span_history){
    register gint i;
    register SDP_SpanData *span_data;
    for(i = 0; i < span_history->model->span_list->len; i++){
        span_data = span_history->span_data[i];
        SDP_SpanData_destroy(span_data);
        }
    C4_Model_destroy(span_history->model);
    g_free(span_history->span_data);
    g_free(span_history);
    return;
    }

static SDP_SpanData *SDP_SpanHistory_get_span_data(
                                         SDP_Scheduler *scheduler,
                                         C4_State *state){
    register C4_Span *span;
    span = scheduler->span_map[state->id];
    return span?scheduler->span_history->span_data[span->id]:NULL;
    }

/**/

static SDP_Cell *SDP_Cell_create(gint query_pos,
                                 SDP_Scheduler *scheduler){
    register SDP_Cell *cell = RecycleBin_alloc(scheduler->cell_recycle);
    register C4_Model *model = scheduler->sdp_pair->sdp->model;
    register gint i, score_width = 3; /* score, max, extra */
    if(scheduler->is_forward){
        g_assert(query_pos >= 0);
    } else {
        g_assert(query_pos <= 0);
        }
    cell->query_pos = query_pos;
    if(scheduler->is_forward && scheduler->sdp_pair->sdp->use_boundary){
        score_width += 2; /* query_start, target_start */
        if(scheduler->traceback){
            score_width += 2; /* query_entry, target_entry */
            }
        }
    cell->score = (C4_Score**)Matrix2d_create(model->state_list->len,
              score_width + model->shadow_list->len, sizeof(C4_Score));
    for(i = 0; i < model->state_list->len; i++)
        cell->score[i][0] = C4_IMPOSSIBLY_LOW_SCORE;
    /* FIXME: implement cell memory managment properly:
     *        (using recyclebins for scores)
     */
    return cell;
    }

static void SDP_Cell_destroy(SDP_Cell *cell, SDP_Scheduler *scheduler){
    g_free(cell->score);
    RecycleBin_recycle(scheduler->cell_recycle, cell);
    return;
    }

static void SDP_Cell_shadow_start(C4_Transition *transition,
                              C4_Score *dst_cell, gint shadow_start,
                              gint dst_query_pos, gint dst_target_pos,
                              gpointer user_data){
    register gint i;
    register C4_Shadow *shadow;
    for(i = 0; i < transition->input->src_shadow_list->len; i++){
        shadow = transition->input->src_shadow_list->pdata[i];
        dst_cell[shadow_start+shadow->designation]
              = shadow->start_func(
                dst_query_pos, dst_target_pos, user_data);
        }
    return;
    }

static void SDP_Cell_shadow_end(C4_Transition *transition,
                              C4_Score *src_cell, gint shadow_start,
                              gint dst_query_pos, gint dst_target_pos,
                              gpointer user_data){
    register gint i;
    register C4_Shadow *shadow;
    for(i = 0; i < transition->dst_shadow_list->len; i++){
        shadow = transition->dst_shadow_list->pdata[i];
        shadow->end_func(src_cell[shadow_start+shadow->designation],
                dst_query_pos, dst_target_pos, user_data);
        }
    return;
    }

static void SDP_Cell_assign(SDP_Scheduler *scheduler,
                            C4_Score *src_cell, C4_Score *dst_cell,
                            C4_Score dst_score, C4_Score max_score,
                            C4_Transition *transition, SDP_Seed *seed,
                            gint dst_query_pos, gint dst_target_pos){
    register SDP_SpanData *span_data = NULL;
    register SDP *sdp = scheduler->sdp_pair->sdp;
    register gint i;
    SDP_SpanSeed span_seed;
    if(sdp->model->total_shadow_designations)
        span_seed.shadow_data = g_new0(gint,
                                sdp->model->total_shadow_designations);
    else
        span_seed.shadow_data = NULL;
    dst_cell[0] = dst_score; /* Assign score */
    if(scheduler->traceback){
        dst_cell[2] = transition->id + 1; /* Transition */
    } else {
        dst_cell[2] = seed->seed_id; /* Seed */
        }
    if(scheduler->is_forward){
        if(sdp->use_boundary){
            dst_cell[3] = src_cell[3];
            dst_cell[4] = src_cell[4];
            /* Copy over entry point for traceback */
            if(scheduler->traceback){
                dst_cell[5] = src_cell[5];
                dst_cell[6] = src_cell[6];
                }
            }
        /* Start shadows */
        SDP_Cell_shadow_start(transition, src_cell,
                              scheduler->shadow_start,
                              dst_query_pos-transition->advance_query,
                              dst_target_pos-transition->advance_target,
                              scheduler->sdp_pair->user_data);
        /* Transport shadows */
        for(i = 0; i < sdp->model->total_shadow_designations; i++){
            dst_cell[scheduler->shadow_start+i]
               = src_cell[scheduler->shadow_start+i];
            }
        }
    if(dst_score < max_score){
        dst_cell[1] = max_score;
    } else { /* Is the best score on this path */
        dst_cell[1] = dst_score;
        if(scheduler->is_forward && (!scheduler->traceback)){
            if((transition->output == sdp->model->end_state->state)
            && (seed->max_end->score < dst_score)){
                /* Is best end for this seed */
                seed->max_end->score = dst_score;
                seed->max_end->query_pos = dst_query_pos;
                seed->max_end->target_pos = dst_target_pos;
                if(scheduler->is_forward && sdp->use_boundary){
                    seed->max_start->query_pos = dst_cell[3];
                    seed->max_start->target_pos = dst_cell[4];
                    }
                }
        } else {
            if((transition->input == sdp->model->start_state->state)
            && (seed->max_start->score < dst_score)
            && (!sdp->use_boundary)){
                /* Is best start for this seed */
                seed->max_start->score = dst_score;
                seed->max_start->query_pos = dst_query_pos;
                seed->max_start->target_pos = dst_target_pos;
                }
            }
        }
    if(scheduler->is_forward && sdp->use_boundary){
        /* Copy output to span history */
        span_data = SDP_SpanHistory_get_span_data(scheduler,
                                                  transition->output);
        if(span_data){
            span_seed.score        = dst_cell[0];
            span_seed.max          = dst_cell[1];
            span_seed.seed_id      = dst_cell[2];
            span_seed.query_start  = dst_cell[3];
            span_seed.target_start = dst_cell[4];
            span_seed.query_entry  = dst_query_pos;
            span_seed.target_entry = dst_target_pos;
            /* Freeze shadows */
            for(i = 0; i < sdp->model->total_shadow_designations; i++){
                span_seed.shadow_data[i]
                    = dst_cell[scheduler->shadow_start+i];
                }
            SDP_SpanData_submit(span_data, &span_seed,
                                scheduler->query_cache, scheduler);
            }
        }
    if(span_seed.shadow_data)
        g_free(span_seed.shadow_data);
    return;
    }

/* FIXME: temp */
static gboolean temp_find_func(gint query_pos,
                               gint target_pos,
                               gint path_id,
                               gpointer user_data){
    /*g_message("tff [%d,%d]", query_pos, target_pos); */
    return TRUE;
    }
#if 0
#endif /* 0 */

static void SDP_Cell_process(SDP_Cell *cell, SDP_Row *row,
                             SDP_Scheduler *scheduler){
    register C4_Transition *transition;
    register SDP *sdp = scheduler->sdp_pair->sdp;
    register C4_Score src_score, dst_score, transition_score, max_score;
    register SDP_Row *dst_row;
    register SDP_Cell *dst_cell;
    register SDP_Seed *seed = NULL, *dst_seed;
    register SDP_SpanData *span_data = NULL;
    register gint i, j, seed_id = 0,
                  src_query_pos, src_target_pos,
                  dst_query_pos, dst_target_pos,
                  rel_dst_query_pos, rel_dst_target_pos,
                  input_pos, output_pos;
    Region temp_region; /* FIXME: temp */
    if(scheduler->traceback)
        seed = scheduler->best_seed;
    if(scheduler->is_forward){
        src_query_pos = cell->query_pos;
        src_target_pos = row->target_pos;
    } else {
        src_query_pos = -cell->query_pos;
        src_target_pos = -row->target_pos;
        }
    /* Calculate transitions in reverse order for forward DP */
    for(i = sdp->model->transition_list->len-1; i >= 0; i--){
        transition = sdp->model->transition_list->pdata[i];
        if(C4_Transition_is_span(transition))
            continue;
        if(scheduler->is_forward){
            dst_query_pos = src_query_pos
                          + transition->advance_query;
            dst_target_pos = src_target_pos
                           + transition->advance_target;
            /* FIXME: or should be >= to limit correctly here ? */
            if((dst_query_pos > Region_query_end(scheduler->region))
            || (dst_target_pos > Region_target_end(scheduler->region)))
                continue;
            input_pos = transition->input->id;
            output_pos = transition->output->id;
            rel_dst_query_pos = dst_query_pos;
            rel_dst_target_pos = dst_target_pos;
            if(sdp->use_boundary){
                /* Copy curr best span to input */
                span_data = SDP_SpanHistory_get_span_data(scheduler,
                                                   transition->input);
                if(span_data){
                    SDP_SpanData_get_curr(span_data,
                                          scheduler->query_cache,
                                          cell->query_pos,
                                          row->target_pos);
                    if(span_data->curr_span_seed){
                        cell->score[input_pos][0]
                           = span_data->curr_span_seed->score;
                        cell->score[input_pos][1]
                           = span_data->curr_span_seed->max;
                        cell->score[input_pos][2]
                           = span_data->curr_span_seed->seed_id;
                        cell->score[input_pos][3] = span_data
                            ->curr_span_seed->query_start;
                        cell->score[input_pos][4] = span_data
                            ->curr_span_seed->target_start;
                        /* Harvest entry point for traceback */
                        if(scheduler->traceback){
                            cell->score[input_pos][5] = span_data
                                ->curr_span_seed->query_entry;
                            cell->score[input_pos][6] = span_data
                                ->curr_span_seed->target_entry;
                            }
                        /* Thaw shadows */
                        for(j = 0;
                          j < sdp->model->total_shadow_designations;
                            j++){
                            cell->score[input_pos]
                                       [scheduler->shadow_start+j]
                       = span_data->curr_span_seed->shadow_data[j];
                            }
                        }
                    }
                }
            /* End shadows */
            SDP_Cell_shadow_end(transition,
                                cell->score[input_pos],
                                scheduler->shadow_start,
                                dst_query_pos, dst_target_pos,
                                scheduler->sdp_pair->user_data);
            transition_score = C4_Calc_score(transition->calc,
                        src_query_pos, src_target_pos,
                        scheduler->sdp_pair->user_data);
        } else {
            dst_query_pos = src_query_pos
                          - transition->advance_query;
            dst_target_pos = src_target_pos
                           - transition->advance_target;
            if((dst_query_pos < scheduler->region->query_start)
            || (dst_target_pos < scheduler->region->target_start))
                continue;
            rel_dst_query_pos = -dst_query_pos;
            rel_dst_target_pos = -dst_target_pos;
            input_pos = transition->output->id;
            output_pos = transition->input->id;
            transition_score = C4_Calc_score(transition->calc,
                        dst_query_pos, dst_target_pos,
                        scheduler->sdp_pair->user_data);
            }
        if(!scheduler->traceback){
            seed_id = cell->score[input_pos][2];
            seed = scheduler->sdp_pair->seed_list->pdata[seed_id];
            }
        src_score = cell->score[input_pos][0];
        max_score = cell->score[input_pos][1];
        dst_score = src_score + transition_score;
        if(dst_score < 0)
            continue;
        if((max_score - dst_score)
           > scheduler->sdp_pair->comparison->param->hsp_param->dropoff)
            continue;
        /*
        if(scheduler->is_forward)
          g_message("Calc [%s] at (%d,%d)=>(%d,%d) [%d+%d]=[%d] (m:%d)",
                 transition->name, src_query_pos, src_target_pos,
                 dst_query_pos, dst_target_pos,
                 src_score, transition_score, dst_score, max_score);
        */
/**/
        if(C4_Transition_is_match(transition)){
#if 0
            if(scheduler->subopt_index
            && SubOpt_Index_is_blocked(scheduler->subopt_index,
                        src_query_pos-scheduler->region->query_start))
                continue;
#endif /* 0 */
            /* FIXME: v.temp: should use SubOpt_Index */
            if(scheduler->is_forward)
                Region_init_static(&temp_region,
                     src_query_pos, src_target_pos,
                     transition->advance_query,
                     transition->advance_target);
            else
                Region_init_static(&temp_region,
                     dst_query_pos, dst_target_pos,
                     /* src_query_pos, src_target_pos, */
                     transition->advance_query,
                     transition->advance_target);
            if(scheduler->subopt_index
                && (scheduler->subopt_index->curr_row
                !=  scheduler->subopt_index->blank_row)){
                g_assert(scheduler->subopt_index->curr_row->target_pos
                        ==
                    (src_target_pos-scheduler->region->target_start));
                }
            if(SubOpt_find(scheduler->sdp_pair->subopt, &temp_region,
                           temp_find_func, NULL)){
                /*
                g_message("SubOpt_find %d %d",
                        temp_region.query_start,
                        temp_region.target_start);
                g_assert(SubOpt_Index_is_blocked(
                            scheduler->subopt_index,
                            (scheduler->is_forward
                                     ?src_query_pos
                                     :dst_query_pos)
                                 -scheduler->region->query_start));
                */
                continue;
            } else {
                /*
                if(scheduler->subopt_index){
                g_assert(!SubOpt_Index_is_blocked(
                            scheduler->subopt_index,
                    src_query_pos-scheduler->region->query_start));
                    }
                */
                }
#if 0
#endif /* 0 */
            }
/**/
        /* Fetch/create and set the dst cell */
        g_assert(seed);
        dst_row = Lookahead_get(scheduler->row_index,
                                transition->advance_target);
        if(!dst_row){
            dst_row = SDP_Row_create(rel_dst_target_pos, scheduler);
            Lookahead_set(scheduler->row_index,
                          transition->advance_target, dst_row);
            Lookahead_move(dst_row->cell_index, cell->query_pos);
            }
        g_assert(dst_row);
        g_assert(dst_row->target_pos == rel_dst_target_pos);
        g_assert(dst_row->cell_index->pos == cell->query_pos);
        dst_cell = Lookahead_get(dst_row->cell_index,
                                 transition->advance_query);
        if(dst_cell){
            if(!scheduler->traceback){
                /* Add obstruct when seeds are different
                 * and the score is positive */
                if((seed_id != dst_cell->score[output_pos][2])
                && (dst_cell->score[output_pos][0] >= 0)){
                    dst_seed = scheduler->sdp_pair->seed_list->pdata
                               [dst_cell->score[output_pos][2]];
                    if(scheduler->is_forward){
                        SDP_ObstructList_add(
                              dst_seed->max_end->obstruct_list, seed);
                    } else {
                        SDP_ObstructList_add(
                              dst_seed->max_start->obstruct_list, seed);
                        }
                    }
                }
            g_assert(dst_cell->query_pos == rel_dst_query_pos);
            /* Keep score if higher (don't tie break on max_score) */
            if(dst_score < dst_cell->score[output_pos][0])
                continue;
        } else { /* New cell */
            dst_cell = SDP_Cell_create(rel_dst_query_pos, scheduler);
            Lookahead_set(dst_row->cell_index,
                          transition->advance_query, dst_cell);
            }
        g_assert(dst_cell);
        SDP_Cell_assign(scheduler,
                        cell->score[input_pos],
                        dst_cell->score[output_pos],
                        dst_score, max_score, transition, seed,
                        dst_query_pos, dst_target_pos);
        }
    return;
    }
/* FIXME: tidy and optimise */

/**/

static void SDP_Row_Lookahead_free_func(gpointer data,
                                        gpointer user_data){
    register SDP_Cell *cell = data;
    register SDP_Row *row = user_data;
    g_assert(cell);
    /* Do not destroy cells - just add to the used_list */
    SList_queue(row->used_list, cell);
    return;
    }

static SDP_Row *SDP_Row_create(gint target_pos,
                               SDP_Scheduler *scheduler){
    register SDP_Row *row = g_new0(SDP_Row, 1);
    if(scheduler->is_forward){
        g_assert(target_pos >= 0);
    } else {
        g_assert(target_pos <= 0);
        }
    row->target_pos = target_pos;
    row->cell_index = Lookahead_create(
          scheduler->is_forward?0:-Region_query_end(scheduler->region),
          scheduler->sdp_pair->sdp->model->max_query_advance,
          SDP_Row_Lookahead_free_func, row);
    row->used_list = SList_create(scheduler->slist_set);
    row->unused_list = SList_create(scheduler->slist_set);
    return row;
    }

/**/

static SDP_TracebackCell *SDP_TracebackCell_create(gint query_pos,
               C4_Transition *transition, SDP_Scheduler *scheduler,
               C4_Score *cell_score_list){
    register SDP_TracebackCell *traceback_cell
                 = RecycleBin_alloc(scheduler->traceback_cell_recycle);
    register gint span_total;
    traceback_cell->query_pos = query_pos;
    traceback_cell->transition = transition;
    span_total = scheduler->sdp_pair->sdp->model->span_list->len;
    if(span_total && scheduler->span_map[transition->input->id]){
        /* FIXME: optimisation: alloc should use a RecycleBin */
        traceback_cell->traceback_span = g_new0(SDP_TracebackSpan,
                                                span_total);
        traceback_cell->traceback_span->query_entry
                      = cell_score_list[5];
        traceback_cell->traceback_span->target_entry
                      = cell_score_list[6];
    } else {
        traceback_cell->traceback_span = NULL;
        }
    return traceback_cell;
    }

static void SDP_TracebackCell_destroy(SDP_TracebackCell *traceback_cell,
                                      SDP_Scheduler *scheduler){
    if(traceback_cell->traceback_span)
        g_free(traceback_cell->traceback_span);
        /* FIXME: optimisation : should be freed with RecycleBin */
    RecycleBin_recycle(scheduler->traceback_cell_recycle,
                       traceback_cell);
    return;
    }

/**/

static SDP_TracebackRow *SDP_TracebackRow_create(gint target_pos){
    register SDP_TracebackRow *traceback_row
     = g_new(SDP_TracebackRow, 1);
    traceback_row->target_pos = target_pos;
    traceback_row->traceback_cell_list = g_ptr_array_new();
    return traceback_row;
    }

static void SDP_TracebackRow_destroy(SDP_TracebackRow *traceback_row,
                                     SDP_Scheduler *scheduler){
    register gint i;
    register SDP_TracebackCell *cell;
    for(i = 0; i < traceback_row->traceback_cell_list->len; i++){
        cell = traceback_row->traceback_cell_list->pdata[i];
        SDP_TracebackCell_destroy(cell, scheduler);
        }
    g_ptr_array_free(traceback_row->traceback_cell_list, TRUE);
    g_free(traceback_row);
    return;
    }

static void SDP_TracebackRow_add(SDP_TracebackRow *traceback_row,
                gint query_pos, C4_Transition *transition,
                SDP_Scheduler *scheduler, C4_Score *cell_score_list){
    register SDP_TracebackCell *traceback_cell
        = SDP_TracebackCell_create(query_pos, transition, scheduler,
                                   cell_score_list);
    g_ptr_array_add(traceback_row->traceback_cell_list, traceback_cell);
    return;
    }
/* FIXME: optimisation use RecycleBin for traceback rows and cells */

/**/

static SDP_Traceback *SDP_Traceback_create(void){
    register SDP_Traceback *traceback = g_new(SDP_Traceback, 1);
    traceback->traceback_row_list = g_ptr_array_new();
    return traceback;
    }

static void SDP_Traceback_destroy(SDP_Traceback *traceback,
                                  SDP_Scheduler *scheduler){
    register gint i;
    register SDP_TracebackRow *traceback_row;
    g_assert(traceback->traceback_row_list);
    for(i = 0; i < traceback->traceback_row_list->len; i++){
        traceback_row = traceback->traceback_row_list->pdata[i];
        SDP_TracebackRow_destroy(traceback_row, scheduler);
        }
    g_ptr_array_free(traceback->traceback_row_list, TRUE);
    g_free(traceback);
    return;
    }

static SDP_TracebackRow *SDP_Traceback_add_row(SDP_Traceback *traceback,
                                               gint target_pos){
    register SDP_TracebackRow *traceback_row;
    g_assert(traceback);
    g_assert(traceback->traceback_row_list);
    g_assert(traceback->traceback_row_list);
    traceback_row = SDP_TracebackRow_create(target_pos);
    g_ptr_array_add(traceback->traceback_row_list, traceback_row);
    return traceback_row;
    }

static SDP_TracebackRow *SDP_Traceback_get_last_row(
                         SDP_Traceback *traceback){
    register SDP_TracebackRow *traceback_row;
    g_assert(traceback);
    g_assert(traceback->traceback_row_list);
    g_assert(traceback->traceback_row_list->len);
    traceback_row = traceback->traceback_row_list->pdata
                   [traceback->traceback_row_list->len-1];
    g_assert(traceback_row);
    return traceback_row;
    }

static void SDP_Traceback_remove_empty_last_row(
            SDP_Traceback *traceback, SDP_Scheduler *scheduler){
    register SDP_TracebackRow *traceback_row
           = SDP_Traceback_get_last_row(traceback);
    g_assert(traceback_row);
    if(traceback_row->traceback_cell_list->len)
        return; /* Don't remove if not empty */
    SDP_TracebackRow_destroy(traceback_row, scheduler);
    g_ptr_array_set_size(traceback->traceback_row_list,
                         traceback->traceback_row_list->len-1);
    return;
    }

/**/

static void SDP_Row_traverse_cell_destroy(gpointer data,
                                          gpointer user_data){
    register SDP_Cell *cell = data;
    register SDP_Scheduler *scheduler = user_data;
    register gint i, transition_id, state_id;
    register C4_Transition *transition;
    register C4_State *state;
    register SDP_TracebackRow *traceback_row;
    register Boundary_Row *boundary_row;
    register C4_Model *model = scheduler->sdp_pair->sdp->model;
    if(scheduler->traceback){
        traceback_row
            = SDP_Traceback_get_last_row(scheduler->traceback);
        for(i = 0; i < model->state_list->len; i++){
            state = model->state_list->pdata[i];
            if(cell->score[i][0] < 0)
                continue;
            transition_id = cell->score[i][2];
            if(!transition_id)
                continue;
            transition = model->transition_list->pdata[transition_id-1];
            if(scheduler->is_forward){
                g_assert(state == transition->output);
            } else {
                g_assert(state == transition->input);
                }
            SDP_TracebackRow_add(traceback_row, cell->query_pos,
                                 transition, scheduler,
                                 cell->score[i]);
            }
        }
    if((!scheduler->is_forward) && scheduler->boundary){
        boundary_row = Boundary_get_last_row(scheduler->boundary);
        state_id = model->start_state->state->id;
        if(cell->score[state_id][0] >= 0){
            Boundary_Row_prepend(boundary_row, -cell->query_pos,
                                 cell->score[state_id][2]);
            }
        }
    SDP_Cell_destroy(cell, scheduler);
    return;
    }

static void SDP_Row_destroy(SDP_Row *row, SDP_Scheduler *scheduler){
    register SDP_TracebackRow *traceback_row = NULL;
    register Boundary_Row *boundary_row = NULL;
    if(scheduler->traceback)
        traceback_row = SDP_Traceback_add_row(scheduler->traceback,
                                              row->target_pos);
    if((!scheduler->is_forward) && scheduler->boundary)
        boundary_row = Boundary_add_row(scheduler->boundary,
                                        -row->target_pos);
    Lookahead_destroy(row->cell_index);
    SList_destroy_with_data(row->used_list,
                            SDP_Row_traverse_cell_destroy, scheduler);
    SList_destroy_with_data(row->unused_list,
                            SDP_Row_traverse_cell_destroy, scheduler);
    if(traceback_row)
        SDP_Traceback_remove_empty_last_row(scheduler->traceback,
                                            scheduler);
    if(boundary_row)
        Boundary_remove_empty_last_row(scheduler->boundary);
    g_free(row);
    return;
    }

/* lookahead->pos uses -ve coordinates
 * terminal coordinates are always +ve
 *     cell->query_pos
 * and row->target_pos use reversed coordinates.
 */

static void SDP_Row_add_seed(SDP_Row *row, SDP_Scheduler *scheduler,
                             SDP_SeedInfo *seed_info){
    register SDP_Cell *cell = SDP_Cell_create(seed_info->query_pos,
                                              scheduler);
    register SDP_Cell *first_cell;
    register gint advance;
    if(row->cell_index->mask){ /* Already have cells */
        /* Add to lookahead when within advance range */
        first_cell = Lookahead_get(row->cell_index, 0);
        g_assert(first_cell);
        advance = cell->query_pos - first_cell->query_pos;
        g_assert(advance >= 0);
        if(advance < row->cell_index->max_advance){
            /* Row should not exist already */
            g_assert(!Lookahead_get(row->cell_index, advance));
            Lookahead_set(row->cell_index, advance, cell);
        } else { /* Otherwise add to the unused list */
            SList_queue(row->unused_list, cell);
            }
    } else { /* Row is empty */
        Lookahead_move(row->cell_index, cell->query_pos);
        Lookahead_set(row->cell_index, 0, cell);
        }
    /* FIXME: Should not seed here unless the seed is part of the
     *        update boundary (or the seed is an update seed)
     */
    scheduler->seed_func(scheduler->seed_data, seed_info, cell);
    return;
    }

static gboolean SDP_Row_print_traverse(gpointer data,
                                       gpointer user_data){
    register SDP_Cell *cell = data;
    register gchar *message = user_data;
    g_message("  >[%s] CELL [%d](%p)",
              message, cell->query_pos, cell);
    return FALSE;
    }

static void SDP_Row_print(SDP_Row *row){
    register gint i;
    register SDP_Cell *cell;
    g_print("\n");
    g_message("--[");
    g_message("ROW target_pos [%d] (mask %d) (ma=%d)", row->target_pos,
            row->cell_index->mask, row->cell_index->max_advance);
    g_message("index pos [%d]", row->cell_index->pos);
    for(i = 0; i <= row->cell_index->max_advance; i++){
        cell = Lookahead_get(row->cell_index, i);
        if(cell){
            g_message("  >lookahead (%d) CELL [%d]",
                    i, cell->query_pos);
            }
        }
    SList_traverse(row->unused_list, SDP_Row_print_traverse, "unused");
    SList_traverse(row->used_list, SDP_Row_print_traverse, "used");
    g_message("]--\n");
    return;
    }

typedef struct {
        gint pos;
    gboolean is_valid;
} SDP_Row_Validate;

static gboolean SDP_Row_is_valid_traverse(gpointer data,
                                          gpointer user_data){
    register SDP_Row_Validate *v = user_data;
    register SDP_Cell *cell = data;
    if(cell->query_pos < v->pos){
        v->is_valid = FALSE;
        }
    v->pos = cell->query_pos;
    return FALSE;
    }

static gboolean SDP_Row_is_valid(SDP_Row *row){
    SDP_Row_Validate v;
    register gint i;
    register SDP_Cell *cell;
    v.pos = -987654321;
    v.is_valid = TRUE;
    SList_traverse(row->used_list, SDP_Row_is_valid_traverse, &v);
    for(i = 0; i <= row->cell_index->max_advance; i++){
        cell = Lookahead_get(row->cell_index, i);
        if(cell){
            if(cell->query_pos < v.pos)
                v.is_valid = FALSE;
            }
        }
    SList_traverse(row->unused_list, SDP_Row_is_valid_traverse, &v);
    if(!v.is_valid){
        g_message("Found invalid row");
        SDP_Row_print(row);
        g_error("bad row");
        }
    return v.is_valid;
    }

static void SDP_Row_reset(SDP_Row *row){
    register SList *joined_list;
    Lookahead_reset(row->cell_index);
    /* Prepend anything in used onto unused */
    joined_list = SList_join(row->used_list, row->unused_list);
    row->used_list = row->unused_list;
    row->unused_list = joined_list;
    return;
    }

static void SDP_Row_move(SDP_Row *row, gint query_pos){
    register SDP_Cell *cell;
    register gint advance;
    Lookahead_move(row->cell_index, query_pos);
    /* Move anything in unused list before index onto used list */
    do {
        cell = SList_first(row->unused_list);
        if(!cell)
            break;
        if(cell->query_pos >= query_pos)
            break;
        SList_queue(row->used_list, SList_pop(row->unused_list));
    } while(TRUE);
    /* Move anything in unused list in index into index */
    do {
        cell = SList_first(row->unused_list);
        if(!cell)
            break;
        advance = cell->query_pos - query_pos;
        g_assert(advance >= 0);
        if(advance > row->cell_index->max_advance)
            break;
        Lookahead_set(row->cell_index, advance,
                      SList_pop(row->unused_list));
    } while(TRUE);
    return;
    }

static void SDP_Row_process(SDP_Row *row, SDP_Scheduler *scheduler){
    register SDP_Cell *cell, *next_cell;
    register gint advance;
    /*
    if(!(row->target_pos % 100))
        g_message("process row [%d]", row->target_pos);
    */
    g_assert(SDP_Scheduler_is_valid(scheduler));
    if(scheduler->is_forward)
        SubOpt_Index_set_row(scheduler->subopt_index,
                     row->target_pos-scheduler->region->target_start);
    else
        SubOpt_Index_set_row(scheduler->subopt_index,
                   (-row->target_pos)-scheduler->region->target_start);
    do {
        /* Add first cell if row index is empty */
        if(!row->cell_index->mask){
            if(SList_isempty(row->unused_list))
                break; /* No more cells */
            cell = SList_pop(row->unused_list);
            g_assert(cell);
            Lookahead_move(row->cell_index, cell->query_pos);
            Lookahead_set(row->cell_index, 0, cell);
            }
        /* Populate initial cell positions with seeds
         * (no more than max_query_advance columns)
         */
        cell = Lookahead_get(row->cell_index, 0);
        g_assert(cell);
        do {
            g_assert(row);
            g_assert(row->unused_list);
            next_cell = SList_first(row->unused_list);
            if(!next_cell)
                break;
            advance = next_cell->query_pos - cell->query_pos;
            g_assert(advance >= 0);
            if(advance > row->cell_index->max_advance)
                break;
            /* Cell cannot be already created */
            g_assert(!Lookahead_get(row->cell_index, advance));
            next_cell = SList_pop(row->unused_list);
            Lookahead_set(row->cell_index, advance, next_cell);
        } while(TRUE);
        /* Process cell */
        SDP_Scheduler_align_rows(scheduler);
        SDP_Cell_process(cell, row, scheduler);
        Lookahead_next(row->cell_index);
    } while(TRUE);
    return;
    }

/* FIXME: tidy */

/**/

static void SDP_Scheduler_Lookahead_free_func(gpointer data,
                                              gpointer user_data){
    register SDP_Row *row = data;
    register SDP_Scheduler *scheduler = user_data;
    SDP_Row_destroy(row, scheduler);
    return;
    }

static C4_Span **SDP_Scheduler_get_span_map(C4_Model *model){
    register gint i, j;
    register C4_Span *span,
                **span_map = g_new0(C4_Span*, model->state_list->len);
    register C4_State *state;
    for(i = 0; i < model->state_list->len; i++){
        state = model->state_list->pdata[i];
        for(j = 0; j < model->span_list->len; j++){
            span = model->span_list->pdata[j];
            if(state == span->span_state)
                span_map[state->id] = span;
            }
        }
    return span_map;
    }

static SDP_Scheduler *SDP_Scheduler_create(SDP_Pair *sdp_pair,
                      Region *region,
                      gboolean is_forward, gboolean find_traceback,
                      Boundary *boundary, SDP_Seed *best_seed,
                      SDP_SeedInfo_reset_Func reset_func,
                      SDP_SeedInfo_next_Func next_func,
                      SDP_SeedInfo_get_Func get_func,
                      SDP_SeedInfo_seed_Func seed_func,
                      gpointer seed_data){
    register SDP_Scheduler *scheduler = g_new(SDP_Scheduler, 1);
    scheduler->sdp_pair = sdp_pair;
    scheduler->slist_set = SListSet_create();
    scheduler->is_forward = is_forward;
    /**/
    scheduler->region = Region_share(region);
    scheduler->seed_data = seed_data;
    scheduler->best_seed = best_seed;
    if(find_traceback)
        scheduler->traceback = SDP_Traceback_create();
    else
        scheduler->traceback = NULL;
    scheduler->row_index = Lookahead_create(
               is_forward?0:-Region_target_end(region),
               sdp_pair->sdp->model->max_target_advance,
               SDP_Scheduler_Lookahead_free_func, scheduler);
    scheduler->cell_recycle = RecycleBin_create(
            "SDP_Cell", sizeof(SDP_Cell), 1024);
    scheduler->traceback_cell_recycle = RecycleBin_create(
            "SDP_TracbackCell", sizeof(SDP_TracebackCell), 1024);
    scheduler->shadow_start = 3;
    if(is_forward && sdp_pair->sdp->use_boundary){
        scheduler->shadow_start += 2;
        if(find_traceback)
            scheduler->shadow_start += 2;
        }
    if(sdp_pair->sdp->use_boundary){
        g_assert(boundary);
        scheduler->boundary = Boundary_share(boundary);
        scheduler->query_cache = SDP_Cache_create(
                                   sdp_pair->comparison->query->len+1);
        scheduler->target_cache = SDP_Cache_create(
                                   sdp_pair->comparison->target->len+1);
        scheduler->span_history = SDP_SpanHistory_create(
                                      scheduler->slist_set,
                                      sdp_pair->sdp->model,
                                      scheduler->query_cache);
    } else {
        g_assert(!boundary);
        scheduler->boundary = NULL;
        scheduler->span_history = NULL;
        scheduler->query_cache = NULL;
        scheduler->target_cache = NULL;
        }
    scheduler->reset_func = reset_func;
    scheduler->next_func = next_func;
    scheduler->get_func = get_func;
    scheduler->seed_func = seed_func;
    scheduler->span_map
        = SDP_Scheduler_get_span_map(sdp_pair->sdp->model);
    scheduler->subopt_index = SubOpt_Index_create(sdp_pair->subopt,
                                                  region);
    return scheduler;
    }

static void SDP_Scheduler_destroy(SDP_Scheduler *scheduler){
    if(scheduler->traceback)
        SDP_Traceback_destroy(scheduler->traceback, scheduler);
    if(scheduler->span_history)
        SDP_SpanHistory_destroy(scheduler->span_history);
    if(scheduler->boundary)
        Boundary_destroy(scheduler->boundary);
    Region_destroy(scheduler->region);
    Lookahead_destroy(scheduler->row_index);
    SListSet_destroy(scheduler->slist_set);
    RecycleBin_destroy(scheduler->cell_recycle);
    RecycleBin_destroy(scheduler->traceback_cell_recycle);
    if(scheduler->query_cache)
        SDP_Cache_destroy(scheduler->query_cache);
    if(scheduler->target_cache)
        SDP_Cache_destroy(scheduler->target_cache);
    if(scheduler->span_map)
        g_free(scheduler->span_map);
    if(scheduler->subopt_index)
        SubOpt_Index_destroy(scheduler->subopt_index);
    g_free(scheduler);
    return;
    }

static void SDP_Scheduler_init(SDP_Scheduler *scheduler,
                               SDP_SeedInfo *seed_info){
    /* Load first seed at start of index */
    register SDP_Row *first_row;
    /* Make first row */
    first_row = SDP_Row_create(seed_info->target_pos, scheduler);
    SDP_Row_add_seed(first_row, scheduler, seed_info);
    /* Set row index position */
    Lookahead_move(scheduler->row_index, first_row->target_pos);
    Lookahead_set(scheduler->row_index, 0, first_row);
    return;
    }

static void SDP_Scheduler_add_seed(SDP_Scheduler *scheduler,
                                   SDP_SeedInfo *seed_info){
    register SDP_Row *row, *first_row;
    register gint advance;
    first_row = Lookahead_get(scheduler->row_index, 0);
    g_assert(first_row);
    advance = seed_info->target_pos - first_row->target_pos;
    g_assert(advance >= 0);
    g_assert(advance <= scheduler->row_index->max_advance);
    row = Lookahead_get(scheduler->row_index, advance);
    if(!row){
        row = SDP_Row_create(seed_info->target_pos, scheduler);
        Lookahead_set(scheduler->row_index, advance, row);
        }
    g_assert(row);
    g_assert(row->target_pos == seed_info->target_pos);
    SDP_Row_add_seed(row, scheduler, seed_info);
    return;
    }

static void SDP_Scheduler_align_rows(SDP_Scheduler *scheduler){
    register gint i;
    /* Set other rows in row_index to same position as first row */
    register SDP_Row *first_row, *row;
    register SDP_Cell *first_cell;
    first_row = Lookahead_get(scheduler->row_index, 0);
    g_assert(first_row);
    first_cell = Lookahead_get(first_row->cell_index, 0);
    g_assert(first_cell);
    for(i = 1; i <= scheduler->row_index->max_advance; i++){
        row = Lookahead_get(scheduler->row_index, i);
        if(row)
            SDP_Row_move(row, first_cell->query_pos);
        }
    /* FIXME: optimisation
     *        can do this faster with Lookahead_next()
     */
    return;
    }

static void SDP_Scheduler_reset_rows(SDP_Scheduler *scheduler){
    register gint i;
    register SDP_Row *row;
    for(i = 0; i <= scheduler->row_index->max_advance; i++){
        row = Lookahead_get(scheduler->row_index, i);
        if(row){
            g_assert(SDP_Row_is_valid(row)); /* FIXME: temp */
            SDP_Row_reset(row);
            }
        }
    return;
    }

static gboolean SDP_Scheduler_is_valid(SDP_Scheduler *scheduler){
    register gint i;
    register SDP_Row *row;
    for(i = 0; i <= scheduler->row_index->max_advance; i++){
        row = Lookahead_get(scheduler->row_index, i);
        g_assert((!row) || SDP_Row_is_valid(row));
        }
    return TRUE;
    }

/**/

typedef struct {
    C4_Transition *transition;
             gint  length;
} SDP_TracebackOperation;

typedef struct {
    GPtrArray *operation_list;
} SDP_TracebackList;

/**/

static SDP_TracebackOperation *SDP_TracebackOperation_create(
                               C4_Transition *transition,
                               gint length){
    register SDP_TracebackOperation *operation
        = g_new(SDP_TracebackOperation, 1);
    /* FIXME: optimisation : use RecycleBin */
    operation->transition = transition;
    operation->length = length;
    return operation;
    }

static void SDP_TracebackOperation_destroy(
            SDP_TracebackOperation *operation){
    g_free(operation);
    return;
    }

/**/

static SDP_TracebackList *SDP_TracebackList_create(void){
    register SDP_TracebackList *traceback_list
                 = g_new(SDP_TracebackList, 1);
    traceback_list->operation_list = g_ptr_array_new();
    return traceback_list;
    }

static void SDP_TracebackList_destroy(
            SDP_TracebackList *traceback_list){
    register gint i;
    register SDP_TracebackOperation *operation;
    for(i = 0; i < traceback_list->operation_list->len; i++){
        operation = traceback_list->operation_list->pdata[i];
        SDP_TracebackOperation_destroy(operation);
        }
    g_ptr_array_free(traceback_list->operation_list, TRUE);
    g_free(traceback_list);
    return;
    }

static void SDP_TracebackList_add(SDP_TracebackList *traceback_list,
                                  C4_Transition *transition,
                                  gint length){
    /* FIXME: optimisation: append to prev (when possible) */
    register SDP_TracebackOperation *operation
        = SDP_TracebackOperation_create(transition, length);
    g_ptr_array_add(traceback_list->operation_list, operation);
    return;
    }

/**/

static void SDP_TracebackList_add_span(
            SDP_TracebackList *traceback_list, C4_Span *span,
            gint query_len, gint target_len){
    g_assert(span);
    g_assert(span->query_loop || (!query_len));
    g_assert(span->target_loop || (!target_len));
    if(query_len){
        g_assert(query_len >= 0);
        g_assert(span->query_loop);
        SDP_TracebackList_add(traceback_list, span->query_loop,
                              query_len);
        }
    if(target_len){
        g_assert(target_len >= 0);
        g_assert(span->target_loop);
        SDP_TracebackList_add(traceback_list, span->target_loop,
                              target_len);
        }
    return;
    }

static SDP_TracebackList *SDP_Scheduler_get_traceback_list(
                  SDP_Scheduler *scheduler){
    register SDP_TracebackRow *traceback_row;
    register SDP_TracebackCell *traceback_cell;
    register gint i, query_pos, target_pos, row_pos, cell_pos;
    register C4_Transition *transition;
    register SDP_TracebackList *traceback_list
           = SDP_TracebackList_create();
    register C4_Model *model = scheduler->sdp_pair->sdp->model;
    g_assert(scheduler);
    g_assert(scheduler->traceback);
    g_assert(scheduler->traceback->traceback_row_list);
    g_assert(scheduler->traceback->traceback_row_list->len);
    row_pos = scheduler->traceback->traceback_row_list->len-1;
    traceback_row = scheduler->traceback->traceback_row_list->pdata
                                                             [row_pos];
    g_assert(traceback_row);
    g_assert(traceback_row->traceback_cell_list->len);
    cell_pos = traceback_row->traceback_cell_list->len-1;
    traceback_cell
              = traceback_row->traceback_cell_list->pdata[cell_pos];
    g_assert(traceback_cell);
    if(scheduler->is_forward){
        g_assert(scheduler->best_seed->max_end->query_pos
                 == traceback_cell->query_pos);
        g_assert(scheduler->best_seed->max_end->target_pos
                 == traceback_row->target_pos);
    } else {
        g_assert(scheduler->best_seed->max_start->query_pos
                 == -traceback_cell->query_pos);
        g_assert(scheduler->best_seed->max_start->target_pos
                 == -traceback_row->target_pos);
        }
    traceback_cell = NULL;
    /* Find the transition to the end state */
    for(i = cell_pos; i >= 0; i--){
        traceback_cell = traceback_row->traceback_cell_list->pdata[i];
        if(scheduler->is_forward){
            if(traceback_cell->transition->output
            == model->end_state->state)
                break;
        } else {
            if(traceback_cell->transition->input
            == model->start_state->state)
                break;
            }
        traceback_cell = NULL;
        }
    g_assert(traceback_cell);
    transition = traceback_cell->transition;
    SDP_TracebackList_add(traceback_list, transition, 1);
    /* Find other transitions */
    do {
        if(traceback_cell->traceback_span){
            query_pos = traceback_cell->traceback_span->query_entry;
            target_pos = traceback_cell->traceback_span->target_entry;
            /* Add span to transition list */
            SDP_TracebackList_add_span(traceback_list,
                    scheduler->span_map[transition->input->id],
                    traceback_cell->query_pos - query_pos
                    - traceback_cell->transition->advance_query,
                    traceback_row->target_pos - target_pos
                    - traceback_cell->transition->advance_target);
        } else {
            query_pos = traceback_cell->query_pos
                      - traceback_cell->transition->advance_query;
            target_pos = traceback_row->target_pos
                       - traceback_cell->transition->advance_target;
            }
        /* Move to the correct row */
        while(traceback_row->target_pos != target_pos){
            g_assert(row_pos >= 0);
            traceback_row
                = scheduler->traceback->traceback_row_list->pdata
                                                           [row_pos--];
            }
        g_assert(traceback_row->target_pos == target_pos);
        /* Move to the correct cell and state */
        cell_pos = traceback_row->traceback_cell_list->len-1;
        traceback_cell
            = traceback_row->traceback_cell_list->pdata[cell_pos];
        if(scheduler->is_forward){
            while((traceback_cell->query_pos != query_pos)
            || (traceback_cell->transition->output
                != transition->input)){
                g_assert(cell_pos >= 0);
                traceback_cell = traceback_row
                               ->traceback_cell_list->pdata[cell_pos--];
                }
        } else {
            while((traceback_cell->query_pos != query_pos)
            || (traceback_cell->transition->input
                != transition->output)){
                g_assert(cell_pos >= 0);
                traceback_cell = traceback_row
                               ->traceback_cell_list->pdata[cell_pos--];
                }
            }
        g_assert(traceback_cell->query_pos == query_pos);
        if(scheduler->is_forward){
            g_assert(traceback_cell->transition->output
                  == transition->input);
        } else {
            g_assert(traceback_cell->transition->input
                    == transition->output);
            }
        transition = traceback_cell->transition;
        SDP_TracebackList_add(traceback_list, transition, 1);
    } while(scheduler->is_forward /* FIXME: tidy */
     ?(transition->input != model->start_state->state)
     :(transition->output != model->end_state->state));
    /* FIXME: g_assert() current position is at max_start */
    return traceback_list;
    }

static Alignment *SDP_Scheduler_add_to_alignment(
                  SDP_Scheduler *scheduler,
                  Alignment *alignment,
                  SDP_TracebackList *traceback_list){
    g_assert(scheduler);
    register gint i, skip;
    register SDP_TracebackOperation *operation;
    if(scheduler->is_forward){
        /* Skip overlap transition for bidirectional traceback */
        skip = scheduler->sdp_pair->sdp->use_boundary?0:1;
        /* FIXME: bug: change skip behaviour for operations */
        for(i = traceback_list->operation_list->len-1-skip;
            i >= 0; i--){
            operation = traceback_list->operation_list->pdata[i];
            Alignment_add(alignment, operation->transition,
                                     operation->length);
            }
    } else {
        for(i = 0; i < (traceback_list->operation_list->len-1); i++){
            operation = traceback_list->operation_list->pdata[i];
            Alignment_add(alignment, operation->transition,
                                     operation->length);
            }
        }
    return alignment;
    }

/**/

static int SDP_Seed_compare_cobs_in_forward_dp_order(const void *a,
                                                     const void *b){
    register SDP_Seed **seed_a = (SDP_Seed**)a,
                      **seed_b = (SDP_Seed**)b;
    register gint target_diff = HSP_target_cobs((*seed_a)->hsp)
                              - HSP_target_cobs((*seed_b)->hsp);
    if(!target_diff)
        return HSP_query_cobs((*seed_a)->hsp)
             - HSP_query_cobs((*seed_b)->hsp);
    return target_diff;
    }

static gboolean SDP_Pair_PQ_compare(gpointer a, gpointer b,
                                    gpointer user_data){
    register SDP_Seed *seed_a = a, *seed_b = b;
    return seed_a->max_end->score
         > seed_b->max_end->score;
    }

SDP_Pair *SDP_Pair_create(SDP *sdp, Comparison *comparison,
                          gpointer user_data){
    register SDP_Pair *sdp_pair = g_new(SDP_Pair, 1);
    register gint i;
    register SDP_Seed *seed;
    g_assert(sdp);
    sdp_pair->sdp = SDP_share(sdp);
    g_assert(comparison);
    g_assert(comparison->hspset);
    g_assert(comparison->hspset->is_finalised);
    g_assert(comparison->hspset->hsp_list->len);
    sdp_pair->comparison = Comparison_share(comparison);
    sdp_pair->user_data = user_data;
    sdp_pair->alignment_count = 0;
    sdp_pair->subopt = SubOpt_create(comparison->query->len,
                                     comparison->target->len);
    /* Build seed list */
    sdp_pair->seed_list = g_ptr_array_new();
    g_ptr_array_set_size(sdp_pair->seed_list,
                         comparison->hspset->hsp_list->len);
    for(i = 0; i < comparison->hspset->hsp_list->len; i++){
        seed = SDP_Seed_create(comparison->hspset->hsp_list->pdata[i]);
        sdp_pair->seed_list->pdata[i] = seed;
        }
    /* Sort seeds on cobs point in DP order */
    g_assert(sdp_pair->seed_list->len);
    qsort(sdp_pair->seed_list->pdata,
          sdp_pair->seed_list->len, sizeof(gpointer),
          SDP_Seed_compare_cobs_in_forward_dp_order);
    /* Assign seed_ids in cobs point DP order */
    for(i = 0; i < sdp_pair->seed_list->len; i++){
        seed = sdp_pair->seed_list->pdata[i];
        seed->seed_id = i;
        }
    sdp_pair->boundary = NULL;
    sdp_pair->seed_pq = PQueue_create(sdp->pq_set,
                                      SDP_Pair_PQ_compare, NULL);
    sdp_pair->last_reported = NULL;
    return sdp_pair;
    }

void SDP_Pair_destroy(SDP_Pair *sdp_pair){
    register gint i;
    register SDP_Seed *seed;
    g_assert(sdp_pair);
    Comparison_destroy(sdp_pair->comparison);
    SDP_destroy(sdp_pair->sdp);
    for(i = 0; i < sdp_pair->seed_list->len; i++){
        seed = sdp_pair->seed_list->pdata[i];
        SDP_Seed_destroy(seed);
        }
    g_ptr_array_free(sdp_pair->seed_list, TRUE);
    SubOpt_destroy(sdp_pair->subopt);
    if(sdp_pair->boundary)
        Boundary_destroy(sdp_pair->boundary);
    PQueue_destroy(sdp_pair->seed_pq, NULL, NULL);
    g_free(sdp_pair);
    return;
    }

/**/

static void SDP_Pair_calculate(SDP_Scheduler *scheduler){
    SDP_SeedInfo seed_info;
    register SDP_Row *row;
    register gint i;
    register C4_Calc *calc;
    register SDP *sdp = scheduler->sdp_pair->sdp;
    if(sdp->model->init_func)
        sdp->model->init_func(scheduler->region,
                              scheduler->sdp_pair->user_data);
    for(i = 0; i < sdp->model->calc_list->len; i++){
        calc = sdp->model->calc_list->pdata[i];
        if(calc && calc->init_func)
            calc->init_func(scheduler->region,
                            scheduler->sdp_pair->user_data);
        }
    scheduler->reset_func(scheduler->seed_data);
    /* Compute seeded DP rows while rows and seeds remain */
    do {
        /* Add first point if row_index is empty */
        if(!scheduler->row_index->mask){
            if(!scheduler->get_func(scheduler->seed_data, &seed_info))
                break;
            SDP_Scheduler_init(scheduler, &seed_info);
            scheduler->next_func(scheduler->seed_data);
            }
        /* Populate initial DP rows with seeds
         * (no more than max_target_advance rows)
         */
        row = Lookahead_get(scheduler->row_index, 0);
        g_assert(row);
        while(scheduler->get_func(scheduler->seed_data, &seed_info)){
            if((seed_info.target_pos - row->target_pos)
              > sdp->model->max_target_advance)
                break;
            SDP_Scheduler_add_seed(scheduler, &seed_info);
            scheduler->next_func(scheduler->seed_data);
            }
        SDP_Scheduler_reset_rows(scheduler);
        SDP_Row_process(row, scheduler);
        Lookahead_next(scheduler->row_index);
    } while(TRUE);
    for(i = 0; i < sdp->model->calc_list->len; i++){
        calc = sdp->model->calc_list->pdata[i];
        if(calc && calc->exit_func)
            calc->exit_func(scheduler->region,
                            scheduler->sdp_pair->user_data);
        }
    if(sdp->model->exit_func)
        sdp->model->exit_func(scheduler->region,
                              scheduler->sdp_pair->user_data);
    return;
    }

/**/

typedef struct {
     GPtrArray *seed_list;
          gint  seed_pos;
          gint  seed_state_id;
} SDP_SeedInfo_List;

static SDP_SeedInfo_List *SDP_SeedInfo_List_create(
                          GPtrArray *seed_list,
                          gboolean is_forward, C4_Model *model){
    register SDP_SeedInfo_List *seed_info_list
        = g_new(SDP_SeedInfo_List, 1);
    seed_info_list->seed_list = seed_list;
    seed_info_list->seed_pos = 0;
    if(is_forward){
        seed_info_list->seed_state_id = model->start_state->state->id;
    } else {
        seed_info_list->seed_state_id = model->end_state->state->id;
        }
    return seed_info_list;
    }

static void SDP_SeedInfo_List_destroy(
            SDP_SeedInfo_List *seed_info_list){
    g_free(seed_info_list);
    return;
    }

static void SDP_SeedInfo_List_reset_forward(gpointer seed_data){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    g_assert(seed_info_list);
    seed_info_list->seed_pos = 0;
    return;
    }

static void SDP_SeedInfo_List_reset_reverse(gpointer seed_data){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    g_assert(seed_info_list);
    seed_info_list->seed_pos = seed_info_list->seed_list->len - 1;
    return;
    }

static void SDP_SeedInfo_List_next_forward(gpointer seed_data){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    g_assert(seed_info_list->seed_pos < seed_info_list->seed_list->len);
    seed_info_list->seed_pos++;
    return;
    }

static void SDP_SeedInfo_List_next_reverse(gpointer seed_data){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    g_assert(seed_info_list->seed_pos >= 0);
    seed_info_list->seed_pos--;
    return;
    }

static gboolean SDP_SeedInfo_List_get_forward(gpointer seed_data,
                                      SDP_SeedInfo *seed_info){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    register SDP_Seed *seed;
    if(seed_info_list->seed_pos >= seed_info_list->seed_list->len)
        return FALSE;
    seed = seed_info_list->seed_list->pdata[seed_info_list->seed_pos];
    seed_info->query_pos = HSP_query_cobs(seed->hsp);
    seed_info->target_pos = HSP_target_cobs(seed->hsp);
    seed_info->seed_id = seed->seed_id;
    seed_info->start_score = seed->max_start->score
                          - (seed->hsp->score >> 1);
    return TRUE;
    }

static gboolean SDP_SeedInfo_List_get_reverse(gpointer seed_data,
                                      SDP_SeedInfo *seed_info){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    register SDP_Seed *seed;
    if(seed_info_list->seed_pos < 0)
        return FALSE;
    seed = seed_info_list->seed_list->pdata[seed_info_list->seed_pos];
    seed_info->query_pos = -HSP_query_cobs(seed->hsp);
    seed_info->target_pos = -HSP_target_cobs(seed->hsp);
    seed_info->seed_id = seed->seed_id;
    seed_info->start_score = (seed->hsp->score >> 1);
    return TRUE;
    }

static void SDP_SeedInfo_List_seed(gpointer seed_data,
                                   SDP_SeedInfo *seed_info,
                                   SDP_Cell *cell){
    register SDP_SeedInfo_List *seed_info_list = seed_data;
    cell->score[seed_info_list->seed_state_id][0]
        = seed_info->start_score;
    cell->score[seed_info_list->seed_state_id][1]
        = seed_info->start_score;
    cell->score[seed_info_list->seed_state_id][2] = seed_info->seed_id;
    return;
    }

/**/

typedef struct {
      Region *region;
    gboolean  is_used;
    SDP_Seed *seed;
        gint  seed_state_id;
} SDP_SeedInfo_Region;

static SDP_SeedInfo_Region *SDP_SeedInfo_Region_create(Region *region,
                  SDP_Seed *seed, gboolean is_forward, C4_Model *model){
    register SDP_SeedInfo_Region *seed_info_region
        = g_new(SDP_SeedInfo_Region, 1);
    seed_info_region->region = Region_share(region);
    seed_info_region->is_used = FALSE;
    seed_info_region->seed = seed;
    if(is_forward){
        seed_info_region->seed_state_id = model->start_state->state->id;
    } else {
        seed_info_region->seed_state_id = model->end_state->state->id;
        }
    return seed_info_region;
    }

static void SDP_SeedInfo_Region_destroy(
            SDP_SeedInfo_Region *seed_info_region){
    Region_destroy(seed_info_region->region);
    g_free(seed_info_region);
    return;
    }

static void SDP_SeedInfo_Region_reset(gpointer seed_data){
    register SDP_SeedInfo_Region *seed_info_region = seed_data;
    g_assert(seed_info_region);
    seed_info_region->is_used = FALSE;
    return;
    }

static void SDP_SeedInfo_Region_next(gpointer seed_data){
    register SDP_SeedInfo_Region *seed_info_region = seed_data;
    g_assert(!seed_info_region->is_used);
    seed_info_region->is_used = TRUE;
    return;
    }

static gboolean SDP_SeedInfo_Region_get_forward(gpointer seed_data,
                                            SDP_SeedInfo *seed_info){
    register SDP_SeedInfo_Region *seed_info_region = seed_data;
    if(seed_info_region->is_used)
        return FALSE;
    seed_info->query_pos = seed_info_region->region->query_start;
    seed_info->target_pos = seed_info_region->region->target_start;
    seed_info->seed_id = seed_info_region->seed->seed_id;
    seed_info->start_score = seed_info_region->seed->max_start->score
                           - (seed_info_region->seed->hsp->score >> 1);
    return TRUE;
    }

#if 0
static gboolean SDP_SeedInfo_Region_get_forward_complete(
                        gpointer seed_data, SDP_SeedInfo *seed_info){
    register SDP_SeedInfo_Region *seed_info_region = seed_data;
    if(seed_info_region->is_used)
        return FALSE;
    seed_info->query_pos = seed_info_region->region->query_start;
    seed_info->target_pos = seed_info_region->region->target_start;
    seed_info->seed_id = seed_info_region->seed->seed_id;
    seed_info->start_score = 0;
    return TRUE;
    }
#endif /* 0 */

static gboolean SDP_SeedInfo_Region_get_reverse(gpointer seed_data,
                                      SDP_SeedInfo *seed_info){
    register SDP_SeedInfo_Region *seed_info_region = seed_data;
    if(seed_info_region->is_used)
        return FALSE;
    seed_info->query_pos
            = -Region_query_end(seed_info_region->region);
    seed_info->target_pos
            = -Region_target_end(seed_info_region->region);
    seed_info->seed_id = seed_info_region->seed->seed_id;
    g_assert(seed_info_region->seed);
    seed_info->start_score = (seed_info_region->seed->hsp->score >> 1);
    return TRUE;
    }

static void SDP_SeedInfo_Region_seed(gpointer seed_data,
                                     SDP_SeedInfo *seed_info,
                                     SDP_Cell *cell){
    register SDP_SeedInfo_Region *seed_info_region = seed_data;
    cell->score[seed_info_region->seed_state_id][0]
        = seed_info->start_score;
    cell->score[seed_info_region->seed_state_id][1]
        = seed_info->start_score;
    cell->score[seed_info_region->seed_state_id][2] = 0;
    /* position 2 is used for traceback data */
    return;
    }

/**/

typedef struct {
         Boundary *boundary;
          gint  curr_row;
          gint  curr_interval;
          gint  curr_pos;
     GPtrArray *seed_list;
      gboolean  is_finished;
          gint  seed_state_id;
} SDP_SeedInfo_Boundary;

static SDP_SeedInfo_Boundary *SDP_SeedInfo_Boundary_create(
                                 Boundary *boundary,
                                 GPtrArray *seed_list,
                                 C4_Model *model){
    register SDP_SeedInfo_Boundary *seed_info_boundary
        = g_new(SDP_SeedInfo_Boundary, 1);
    seed_info_boundary->boundary = Boundary_share(boundary);
    seed_info_boundary->curr_row = 0;
    seed_info_boundary->curr_interval = 0;
    seed_info_boundary->curr_pos = 0;
    seed_info_boundary->seed_list = seed_list;
    seed_info_boundary->is_finished = FALSE;
    seed_info_boundary->seed_state_id = model->start_state->state->id;
    return seed_info_boundary;
    }

static void SDP_SeedInfo_Boundary_destroy(
            SDP_SeedInfo_Boundary *seed_info_boundary){
    Boundary_destroy(seed_info_boundary->boundary);
    g_free(seed_info_boundary);
    return;
    }

static void SDP_SeedInfo_Boundary_reset_forward(gpointer seed_data){
    register SDP_SeedInfo_Boundary *seed_info_boundary = seed_data;
    g_assert(seed_info_boundary);
    seed_info_boundary->curr_row = 0;
    seed_info_boundary->curr_interval = 0;
    seed_info_boundary->curr_pos = 0;
    seed_info_boundary->is_finished = FALSE;
    return;
    }

static void SDP_SeedInfo_Boundary_next_forward(gpointer seed_data){
    register SDP_SeedInfo_Boundary *seed_info_boundary = seed_data;
    register Boundary_Row *boundary_row
        = seed_info_boundary->boundary->row_list->pdata
         [seed_info_boundary->curr_row];
    register Boundary_Interval *interval
        = boundary_row->interval_list->pdata
         [seed_info_boundary->curr_interval];
    /* If more seeds in interval, move curr_pos */
    if(seed_info_boundary->curr_pos < (interval->length-1)){
        seed_info_boundary->curr_pos++;
        return;
        }
    /* If more intervals in row, move curr_interval */
    if(seed_info_boundary->curr_interval
    <  (boundary_row->interval_list->len-1)){
        seed_info_boundary->curr_interval++;
        seed_info_boundary->curr_pos = 0;
        return;
        }
    /* If more rows in boundary, move curr_row */
    if(seed_info_boundary->curr_row
    <  (seed_info_boundary->boundary->row_list->len-1)){
        seed_info_boundary->curr_row++;
        seed_info_boundary->curr_interval = 0;
        seed_info_boundary->curr_pos = 0;
        return;
        }
    seed_info_boundary->is_finished = TRUE;
    return;
    }

static gboolean SDP_SeedInfo_Boundary_get_forward(gpointer seed_data,
                                          SDP_SeedInfo *seed_info){
    register SDP_SeedInfo_Boundary *seed_info_boundary = seed_data;
    register Boundary_Row *boundary_row
        = seed_info_boundary->boundary->row_list->pdata
         [seed_info_boundary->curr_row];
    register Boundary_Interval *interval
        = boundary_row->interval_list->pdata
         [seed_info_boundary->curr_interval];
    register SDP_Seed *seed;
    if(seed_info_boundary->is_finished)
        return FALSE;
    seed = seed_info_boundary->seed_list->pdata[interval->seed_id];
    seed_info->query_pos = interval->query_pos
                         + seed_info_boundary->curr_pos;
    seed_info->target_pos = boundary_row->target_pos;
    seed_info->seed_id = interval->seed_id;
    seed_info->start_score = 0;
    return TRUE;
    }

static void SDP_SeedInfo_Boundary_seed(gpointer seed_data,
                                       SDP_SeedInfo *seed_info,
                                       SDP_Cell *cell){
    register SDP_SeedInfo_Boundary *seed_info_boundary = seed_data;
    cell->score[seed_info_boundary->seed_state_id][0]
        = seed_info->start_score;
    cell->score[seed_info_boundary->seed_state_id][1]
        = seed_info->start_score;
    cell->score[seed_info_boundary->seed_state_id][2]
        = seed_info->seed_id;
    /* Additionally store start, end pos for seeds */
    cell->score[seed_info_boundary->seed_state_id][3]
        = seed_info->query_pos;
    cell->score[seed_info_boundary->seed_state_id][4]
        = seed_info->target_pos;
    return;
    }

static void SDP_SeedInfo_Boundary_seed_traceback(gpointer seed_data,
                                       SDP_SeedInfo *seed_info,
                                       SDP_Cell *cell){
    register SDP_SeedInfo_Boundary *seed_info_boundary = seed_data;
    cell->score[seed_info_boundary->seed_state_id][0]
        = seed_info->start_score;
    cell->score[seed_info_boundary->seed_state_id][1]
        = seed_info->start_score;
    cell->score[seed_info_boundary->seed_state_id][2]
        = 0; /* Do not set seed_id for traceback */
    /* Additionally store start, end pos for seeds */
    cell->score[seed_info_boundary->seed_state_id][3]
        = seed_info->query_pos;
    cell->score[seed_info_boundary->seed_state_id][4]
        = seed_info->target_pos;
    return;
    }

/**/

static Boundary *SDP_Pair_find_start_points(SDP_Pair *sdp_pair,
                                            GPtrArray *seed_list){
    register SDP_SeedInfo_List *seed_info_list
           = SDP_SeedInfo_List_create(seed_list, FALSE,
                                      sdp_pair->sdp->model);
    register Region *region = Region_create(0, 0,
                        sdp_pair->comparison->query->len,
                        sdp_pair->comparison->target->len);
    register Boundary *boundary = NULL;
    register SDP_Scheduler *scheduler;
    if(sdp_pair->sdp->use_boundary)
        boundary = Boundary_create();
    scheduler = SDP_Scheduler_create(sdp_pair,
                              region, FALSE, FALSE, boundary, NULL,
                              SDP_SeedInfo_List_reset_reverse,
                              SDP_SeedInfo_List_next_reverse,
                              SDP_SeedInfo_List_get_reverse,
                              SDP_SeedInfo_List_seed,
                              seed_info_list);
    SDP_Pair_calculate(scheduler);
    SDP_Scheduler_destroy(scheduler);
    Region_destroy(region);
    SDP_SeedInfo_List_destroy(seed_info_list);
    if(boundary)
        Boundary_reverse(boundary);
    return boundary;
    }

static void SDP_Pair_find_end_points(SDP_Pair *sdp_pair,
                                     GPtrArray *seed_list,
                                     Boundary *boundary){
    register Region *region = Region_create(0, 0,
                        sdp_pair->comparison->query->len,
                        sdp_pair->comparison->target->len);
    register SDP_Scheduler *scheduler;
    register SDP_SeedInfo_List *seed_info_list = NULL;
    register SDP_SeedInfo_Boundary *seed_info_boundary = NULL;
    if(boundary){
        seed_info_boundary = SDP_SeedInfo_Boundary_create(boundary,
                                    seed_list, sdp_pair->sdp->model);
        scheduler = SDP_Scheduler_create(sdp_pair,
                                 region, TRUE, FALSE, boundary, NULL,
                                 SDP_SeedInfo_Boundary_reset_forward,
                                 SDP_SeedInfo_Boundary_next_forward,
                                 SDP_SeedInfo_Boundary_get_forward,
                                 SDP_SeedInfo_Boundary_seed,
                                 seed_info_boundary);
    } else {
        seed_info_list = SDP_SeedInfo_List_create(seed_list, TRUE,
                                                  sdp_pair->sdp->model);
        scheduler = SDP_Scheduler_create(sdp_pair,
                                     region, TRUE, FALSE, NULL, NULL,
                                     SDP_SeedInfo_List_reset_forward,
                                     SDP_SeedInfo_List_next_forward,
                                     SDP_SeedInfo_List_get_forward,
                                     SDP_SeedInfo_List_seed,
                                     seed_info_list);
        }
    SDP_Pair_calculate(scheduler);
    SDP_Scheduler_destroy(scheduler);
    Region_destroy(region);
    if(boundary){
        g_assert(seed_info_boundary);
        SDP_SeedInfo_Boundary_destroy(seed_info_boundary);
    } else {
        g_assert(seed_info_list);
        SDP_SeedInfo_List_destroy(seed_info_list);
        }
    return;
    }

/**/

static GPtrArray *SDP_Pair_merge_seed_list_uniq(GPtrArray *seed_list_a,
                                                GPtrArray *seed_list_b){
    register GPtrArray *seed_list = g_ptr_array_new();
    register gint a = 0, b = 0;
    register SDP_Seed *seed_a, *seed_b, *seed, *last_seed = NULL;
    while((a < seed_list_a->len) && (b < seed_list_b->len)){
        seed_a = seed_list_a->pdata[a];
        seed_b = seed_list_b->pdata[b];
        if(seed_a->seed_id < seed_b->seed_id){
            a++;
            seed = seed_a;
        } else {
            b++;
            seed = seed_b;
            }
        /* Merge uniquely */
        if((!last_seed) || (last_seed->seed_id != seed->seed_id)){
            last_seed = seed;
            g_ptr_array_add(seed_list, seed);
            }
        }
    while(a < seed_list_a->len){
        seed_a = seed_list_a->pdata[a++];
        if((!last_seed) || (last_seed->seed_id != seed_a->seed_id)){
            last_seed = seed_a;
            g_ptr_array_add(seed_list, seed_a);
            }
        }
    while(b < seed_list_b->len){
        seed_b = seed_list_b->pdata[b++];
        if((!last_seed) || (last_seed->seed_id != seed_b->seed_id)){
            last_seed = seed_b;
            g_ptr_array_add(seed_list, seed_b);
            }
        }
    return seed_list;
    }

static Boundary *SDP_Pair_update_starts(SDP_Pair *sdp_pair){
    register gint i;
    register SDP_Seed *seed;
    register GPtrArray *start_seed_list, *end_seed_list, *seed_list;
    /* register GPtrArray *seed_list; */
    register Boundary *boundary = NULL;
    g_assert(sdp_pair->last_reported);
    start_seed_list = SDP_ObstructList_get_list(
                  sdp_pair->last_reported->max_start->obstruct_list);
    end_seed_list = SDP_ObstructList_get_list(
                  sdp_pair->last_reported->max_end->obstruct_list);
    seed_list = SDP_Pair_merge_seed_list_uniq(start_seed_list,
                                              end_seed_list);
    /*
    seed_list = SDP_ObstructList_get_list(
                   sdp_pair->last_reported->max_start->obstruct_list);
    */
    /* g_message("Updating [%d] starts", seed_list->len); */
    /* Set max start scores low */
    for(i = 0; i < seed_list->len; i++){
        seed = seed_list->pdata[i];
        seed->max_start->score = C4_IMPOSSIBLY_LOW_SCORE;
        }
    boundary = SDP_Pair_find_start_points(sdp_pair, seed_list);
    g_ptr_array_free(start_seed_list, TRUE);
    g_ptr_array_free(end_seed_list, TRUE);
    g_ptr_array_free(seed_list, TRUE);
    return boundary;
    }
/* FIXME: revert to combined start+end obstruct lists
 *        as they are no longer treated separately.
 *        (also just need to get the list once for
 *         update starts and ends here)
 */

static void SDP_Pair_update_ends(SDP_Pair *sdp_pair,
                                 Boundary *update_boundary){
    register GPtrArray *start_seed_list, *end_seed_list, *seed_list;
    register gint i;
    register SDP_Seed *seed;
    start_seed_list = SDP_ObstructList_get_list(
                  sdp_pair->last_reported->max_start->obstruct_list);
    end_seed_list = SDP_ObstructList_get_list(
                  sdp_pair->last_reported->max_end->obstruct_list);
    seed_list = SDP_Pair_merge_seed_list_uniq(start_seed_list,
                                              end_seed_list);
    /* g_message("Updating [%d] ends", seed_list->len); */
    /* Set max end scores low */
    for(i = 0; i < seed_list->len; i++){
        seed = seed_list->pdata[i];
        seed->max_end->score = C4_IMPOSSIBLY_LOW_SCORE;
        }
    if(update_boundary){
        /* FIXME: temp */
        if(0)
        Boundary_print(update_boundary, "boundary:update_ends");
        /**/
        SDP_Pair_find_end_points(sdp_pair, sdp_pair->seed_list,
                                 update_boundary);
        /* FIXME:
         * Call SDP_Pair_find_end_points();
         * use update_boundary to recompute ends
         * (allowing landing in full combined boundary,
         *  but only allow seeding in the update_start_boundary)
         * raise changed seeds in the PQueue
         */
    } else {
        SDP_Pair_find_end_points(sdp_pair, seed_list, NULL);
        }
    for(i = 0; i < seed_list->len; i++){
        seed = seed_list->pdata[i];
        if(seed->pq_node){
            if(seed->max_end->score > 0){
                g_assert(seed->pq_node);
                PQueue_raise(sdp_pair->seed_pq, seed->pq_node);
            } else {
                PQueue_remove(sdp_pair->seed_pq, seed->pq_node);
                seed->pq_node = NULL;
                }
            }
        }
    g_ptr_array_free(start_seed_list, TRUE);
    g_ptr_array_free(end_seed_list, TRUE);
    g_ptr_array_free(seed_list, TRUE);
    return;
    }

/**/

static void SDP_Pair_find_pre_cobs_path(SDP_Pair *sdp_pair,
                     SDP_Seed *best_seed, Alignment *alignment){
    register SDP_TracebackList *traceback_list;
    register Region *region
      = Region_create(best_seed->max_start->query_pos,
                      best_seed->max_start->target_pos,
                      HSP_query_cobs(best_seed->hsp)
                    - best_seed->max_start->query_pos,
                      HSP_target_cobs(best_seed->hsp)
                    - best_seed->max_start->target_pos);
    register SDP_SeedInfo_Region *seed_info_region
           = SDP_SeedInfo_Region_create(region, best_seed,
                                        FALSE, sdp_pair->sdp->model);
    register SDP_Scheduler *scheduler;
    scheduler = SDP_Scheduler_create(sdp_pair,
                                     region, FALSE, TRUE, NULL,
                                     best_seed,
                                     SDP_SeedInfo_Region_reset,
                                     SDP_SeedInfo_Region_next,
                                     SDP_SeedInfo_Region_get_reverse,
                                     SDP_SeedInfo_Region_seed,
                                     seed_info_region);
    SDP_Pair_calculate(scheduler);
    traceback_list = SDP_Scheduler_get_traceback_list(scheduler);
    SDP_Scheduler_add_to_alignment(scheduler, alignment,
                                   traceback_list);
    SDP_Scheduler_destroy(scheduler);
    Region_destroy(region);
    SDP_SeedInfo_Region_destroy(seed_info_region);
    SDP_TracebackList_destroy(traceback_list);
    return;
    }

static void SDP_Pair_find_post_cobs_path(SDP_Pair *sdp_pair,
                     SDP_Seed *best_seed, Alignment *alignment){
    register SDP_TracebackList *traceback_list;
    register Region *region
      = Region_create(HSP_query_cobs(best_seed->hsp),
                      HSP_target_cobs(best_seed->hsp),
                      best_seed->max_end->query_pos
                    - HSP_query_cobs(best_seed->hsp),
                      best_seed->max_end->target_pos
                    - HSP_target_cobs(best_seed->hsp));
    register SDP_SeedInfo_Region *seed_info_region
         = SDP_SeedInfo_Region_create(region, best_seed,
                                      TRUE, sdp_pair->sdp->model);
    register SDP_Scheduler *scheduler;
    scheduler = SDP_Scheduler_create(sdp_pair,
                                     region, TRUE, TRUE, NULL,
                                     best_seed,
                                     SDP_SeedInfo_Region_reset,
                                     SDP_SeedInfo_Region_next,
                                     SDP_SeedInfo_Region_get_forward,
                                     SDP_SeedInfo_Region_seed,
                                     seed_info_region);
    SDP_Pair_calculate(scheduler);
    traceback_list = SDP_Scheduler_get_traceback_list(scheduler);
    SDP_Scheduler_add_to_alignment(scheduler, alignment,
                                   traceback_list);
    SDP_Scheduler_destroy(scheduler);
    Region_destroy(region);
    SDP_SeedInfo_Region_destroy(seed_info_region);
    SDP_TracebackList_destroy(traceback_list);
    return;
    }


static void SDP_Pair_find_complete_path(SDP_Pair *sdp_pair,
     SDP_Seed *best_seed, Boundary *boundary, Alignment *alignment){
    register SDP_TracebackList *traceback_list;
    register Region *region
      = Region_create(best_seed->max_start->query_pos,
                      best_seed->max_start->target_pos,
                      best_seed->max_end->query_pos
                    - best_seed->max_start->query_pos,
                      best_seed->max_end->target_pos
                    - best_seed->max_start->target_pos);
    register SDP_SeedInfo_Region *seed_info_region
         = SDP_SeedInfo_Region_create(region, best_seed,
                                      TRUE, sdp_pair->sdp->model);
    register SDP_Scheduler *scheduler;
    register Boundary *sub_boundary = Boundary_select(boundary, region);
    register SDP_SeedInfo_Boundary *seed_info_boundary
        = SDP_SeedInfo_Boundary_create(sub_boundary,
                                       sdp_pair->seed_list,
                                       sdp_pair->sdp->model);
    scheduler = SDP_Scheduler_create(sdp_pair, region, TRUE, TRUE,
                          sub_boundary,
                          best_seed,
                          SDP_SeedInfo_Boundary_reset_forward,
                          SDP_SeedInfo_Boundary_next_forward,
                          SDP_SeedInfo_Boundary_get_forward,
                          SDP_SeedInfo_Boundary_seed_traceback,
                          seed_info_boundary);
    /* FIXME: need sub-boundary here, and to seed starts start only,
     *        but allow landing of spans within boundary regions.
     */
    SDP_Pair_calculate(scheduler);
    traceback_list = SDP_Scheduler_get_traceback_list(scheduler);
    SDP_Scheduler_add_to_alignment(scheduler, alignment,
                                   traceback_list);
    SDP_Scheduler_destroy(scheduler);
    Region_destroy(region);
    SDP_SeedInfo_Region_destroy(seed_info_region);
    Boundary_destroy(sub_boundary);
    SDP_TracebackList_destroy(traceback_list);
    SDP_SeedInfo_Boundary_destroy(seed_info_boundary);
    return;
    }

static Alignment *SDP_Pair_find_path(SDP_Pair *sdp_pair,
                                     SDP_Seed *best_seed,
                                     Boundary *boundary){
    register Region *alignment_region
           = Region_create(best_seed->max_start->query_pos,
                           best_seed->max_start->target_pos,
                           best_seed->max_end->query_pos
                          -best_seed->max_start->query_pos,
                           best_seed->max_end->target_pos
                          -best_seed->max_start->target_pos);
    register Alignment *alignment = Alignment_create(
                         sdp_pair->sdp->model,
                         alignment_region, best_seed->max_end->score);
    if(sdp_pair->sdp->use_boundary){
        g_assert(boundary);
        SDP_Pair_find_complete_path(sdp_pair, best_seed,
                                    boundary, alignment);
    } else {
        SDP_Pair_find_pre_cobs_path(sdp_pair, best_seed, alignment);
        SDP_Pair_find_post_cobs_path(sdp_pair, best_seed, alignment);
        }
    Region_destroy(alignment_region);
    g_assert(Alignment_is_valid(alignment, alignment_region,
                                sdp_pair->user_data));
    return alignment;
    }

static void SDP_Pair_fill_pqueue(SDP_Pair *sdp_pair){
    register gint i;
    register SDP_Seed *seed;
    for(i = 0; i < sdp_pair->seed_list->len; i++){
        seed = sdp_pair->seed_list->pdata[i];
        if(seed->max_end->score >= 0){
            g_assert(!seed->pq_node);
            seed->pq_node = PQueue_push(sdp_pair->seed_pq, seed);
            }
        }
    return;
    }

Alignment *SDP_Pair_next_path(SDP_Pair *sdp_pair, C4_Score threshold){
    register SDP_Seed *best_seed = NULL;
    register Alignment *alignment;
    register Boundary *update_boundary;
    /* Make the start and end points up to date */
    if(sdp_pair->alignment_count){
        update_boundary = SDP_Pair_update_starts(sdp_pair);
        if(sdp_pair->boundary
           && sdp_pair->boundary->row_list->len
           && update_boundary
           && update_boundary->row_list->len)
            Boundary_insert(sdp_pair->boundary, update_boundary);
        SDP_Pair_update_ends(sdp_pair, sdp_pair->boundary);
        if(update_boundary)
            Boundary_destroy(update_boundary);
    } else {
        /*
        g_message("find [%d] starts", sdp_pair->seed_list->len);
        */
        sdp_pair->boundary = SDP_Pair_find_start_points(sdp_pair,
                                              sdp_pair->seed_list);
        /*
        g_message("find [%d] ends", sdp_pair->seed_list->len);
        */
        SDP_Pair_find_end_points(sdp_pair, sdp_pair->seed_list,
                                 sdp_pair->boundary);
        /*
        g_message("have ends");
        */
        SDP_Pair_fill_pqueue(sdp_pair);
        }
    /* Find best seed */
    best_seed = PQueue_pop(sdp_pair->seed_pq);
    if(!best_seed)
        return NULL;
    best_seed->pq_node = NULL;
    /*
    g_message("best_score [%d](id=%d) [%d,%d]->[%d,%d]",
               best_seed->max_end->score,
               best_seed->seed_id,
               best_seed->max_start->query_pos,
               best_seed->max_start->target_pos,
               best_seed->max_end->query_pos,
               best_seed->max_end->target_pos);
    */
    if(best_seed->max_end->score < threshold)
        return NULL; /* Score below threshold */
    sdp_pair->alignment_count++;
    sdp_pair->last_reported = best_seed;
    alignment = SDP_Pair_find_path(sdp_pair, best_seed,
                                   sdp_pair->boundary);
    /* Remove the best seed and add alignment to SubOpt */
    best_seed->max_end->score = C4_IMPOSSIBLY_LOW_SCORE;
    SubOpt_add_alignment(sdp_pair->subopt, alignment);
    /*
    g_message("have alignment [%d]", alignment->score);
    g_message("Have [%d] spans", sdp_pair->sdp->model->span_list->len);
    */
    return alignment;
    }

/* Suboptimal alignment algorithm:
 * ------------------------------
 * take start + end obstruct list of prev_reported_alignment_seed
 * build update_start_boundary (applying subopt)
 *
 * use update_start_boundary to recompute ends
 * (allowing landing in full combined boundary,
 * but only allow seeding in the update_start_boundary)
 * raise changed seeds in the PQueue
 *
 * PQ pop and report best
 */

/**/

/* FIXME: split seeded viterbi code (SDP_Scheduler etc)
 *        to a separtate module
 */

