/****************************************************************\
*                                                                *
*  UGAM: UnGapped Alignment Manager                              *
*                                                                *
*  Guy St.C. Slater..   mailto:guy@ebi.ac.uk                     *
*  Copyright (C) 2000-2005.  All Rights Reserved.                *
*                                                                *
*  This source code is distributed under the terms of the        *
*  GNU Lesser General Public License. See the file COPYING       *
*  or http://www.fsf.org/copyleft/lesser.html for details        *
*                                                                *
*  If you use this code, please keep this notice intact.         *
*                                                                *
\****************************************************************/

#include <string.h> /* For strlen()  */
#include <math.h>   /* For pow()     */
#include <ctype.h>  /* For toupper() */
#include "ugam.h"

/**/

static gint UGAM_Matcher_get_expect(UGAM *ugam, gint len,
                                    UGAM_StrandParam *strand_param){
    register gint expect;
    register gdouble expectation;
    /* FIXME: should just calculate expectation once for the UGAM */
    expectation = 1.0/pow(strand_param->alphabet_size,
                          strand_param->wordlen);
    expect = (gint)((expectation * (len - strand_param->wordlen+1))
                  + ugam->saturate_threshold);
    return expect;
    }

static UGAM_Matcher_QueryInfo *UGAM_Matcher_QueryInfo_create(
                               UGAM *ugam, Sequence *query){
    register UGAM_Matcher_QueryInfo *query_info
         = g_new(UGAM_Matcher_QueryInfo, 1);
    query_info->query = Sequence_share(query);
    query_info->curr_comparison = NULL; /* Set later */
    return query_info;
    }

static void UGAM_Matcher_QueryInfo_destroy(
            UGAM_Matcher_QueryInfo *query_info){
    g_assert(query_info);
    Sequence_destroy(query_info->query);
    if(query_info->curr_comparison)
        Comparison_destroy(query_info->curr_comparison);
    g_free(query_info);
    return;
    }

/**/

/* FIXME: implement variable sizes for WordInfo and QPos */

static UGAM_Matcher_WordInfo *UGAM_Matcher_WordInfo_create(
                              UGAM_Matcher *matcher){
    register UGAM_Matcher_WordInfo *wordinfo
        = RecycleBin_alloc(matcher->recycle_wordinfo);
    wordinfo->query_pos_list = NULL;
    wordinfo->wordhood_list = NULL;
    wordinfo->match_count = 0;
    wordinfo->match_mailbox = -1;
    return wordinfo;
    }

static void UGAM_Matcher_WordInfo_empty(UGAM_Matcher *matcher,
            UGAM_Matcher_WordInfo *wordinfo){
    register UGAM_Matcher_QPos *qpos;
    register UGAM_Matcher_NPos *npos;
    while(wordinfo->query_pos_list){
        qpos = wordinfo->query_pos_list;
        wordinfo->query_pos_list = qpos->next;
        RecycleBin_recycle(matcher->recycle_qpos, qpos);
        }
    while(wordinfo->wordhood_list){
        npos = wordinfo->wordhood_list;
        wordinfo->wordhood_list = npos->next;
        RecycleBin_recycle(matcher->recycle_npos, npos);
        }
    return;
    }

static void UGAM_Matcher_WordInfo_add_qpos(UGAM *ugam,
            UGAM_Matcher *matcher,
            UGAM_Matcher_WordInfo *wordinfo,
            UGAM_Matcher_QueryInfo *query_info, gint pos){
    register UGAM_Matcher_QPos *qpos;
    g_assert(wordinfo);
    if(!wordinfo->match_mailbox) /* Already blocked */
        return;
    /* Apply saturate_threshold to frequent query words */
    if((ugam->saturate_threshold)
    && (++wordinfo->match_count > matcher->query_expect)){
        wordinfo->match_mailbox = 0; /* block */
        UGAM_Matcher_WordInfo_empty(matcher, wordinfo);
        return;
        }
    qpos = RecycleBin_alloc(matcher->recycle_qpos);
    qpos->next = wordinfo->query_pos_list;
    qpos->query_pos = pos;
    qpos->query_info = query_info;
    wordinfo->query_pos_list = qpos;
    return;
    }

static void UGAM_Matcher_WordInfo_add_npos(UGAM_Matcher *matcher,
            UGAM_Matcher_WordInfo *wordinfo,
            UGAM_Matcher_WordInfo *neighbour){
    register UGAM_Matcher_NPos *npos;
    g_assert(wordinfo);
    g_assert(neighbour);
    if(!wordinfo->match_mailbox) /* Already blocked */
        return;
    npos = RecycleBin_alloc(matcher->recycle_npos);
    npos->next = wordinfo->wordhood_list;
    npos->neighbour = neighbour;
    wordinfo->wordhood_list = npos;
    return;
    }

typedef struct {
         UGAM *ugam;
     Sequence *target;
         gint  curr_frame;
    GPtrArray *queryinfo_list; /* queryies containing hspsets */
         gint  target_expect;
} UGAM_Matcher_TargetInfo;

static void UGAM_Matcher_WordInfo_seed(
            UGAM_Matcher_QueryInfo *query_info,
            UGAM_Matcher_TargetInfo *target_info,
            gint query_pos, gint target_pos){
    g_assert(query_pos >= 0);
    g_assert(target_pos >= 0);
    if(!query_info->curr_comparison){
        if(target_info->ugam->scan_query)
            query_info->curr_comparison = Comparison_create(
                target_info->ugam->comparison_param->mirror,
                target_info->target, query_info->query);
        else
            query_info->curr_comparison = Comparison_create(
                target_info->ugam->comparison_param,
                query_info->query, target_info->target);
        g_ptr_array_add(target_info->queryinfo_list, query_info);
        target_info->ugam->curr_matcher->comparison_count++;
        if(target_info->ugam->saturate_threshold)
           target_info->target_expect = UGAM_Matcher_get_expect(
                    target_info->ugam,
                    target_info->target->len,
                    target_info->ugam->param->target_param);
        else
           target_info->target_expect = 0;
        }
    if(target_info->ugam->scan_query){
        HSPset_seed_hsp(query_info->curr_comparison->hspset,
               target_pos, query_pos,
               target_info->ugam->seed_len);
    } else {
        HSPset_seed_hsp(query_info->curr_comparison->hspset,
               query_pos, target_pos, target_info->ugam->seed_len);
        }
    return;
    }

static void UGAM_Matcher_WordInfo_traverse(
            UGAM_Matcher_WordInfo *wordinfo,
            UGAM_Matcher_TargetInfo *target_info, gint target_pos){
    register UGAM_Matcher_QPos *qpos;
    register UGAM_Matcher_NPos *npos;
    register UGAM *ugam = target_info->ugam;
    g_assert(target_pos >= 0);
    /* Ignore if over saturate expectation */
    if(ugam->saturate_threshold){
        if(!wordinfo->match_mailbox) /* Already blocked */
            return;
        if(wordinfo->match_mailbox
        == ugam->curr_matcher->comparison_count){
            if(++wordinfo->match_count > target_info->target_expect){
                /*
                g_message("tg word saturated [%d][%.12s]",
                    target_pos, target_info->target->seq+target_pos);
                */
                return; /* Saturated */
                }
        } else {
            wordinfo->match_mailbox
                    = ugam->curr_matcher->comparison_count;
            wordinfo->match_count = 1;
            }
        }
    g_assert(wordinfo->query_pos_list || wordinfo->wordhood_list);
    /* FIXME: obsolete: */
    for(qpos = wordinfo->query_pos_list; qpos; qpos = qpos->next){
        /* FIXME: don't use this query_info if single query */
        UGAM_Matcher_WordInfo_seed(qpos->query_info, target_info,
                                   qpos->query_pos, target_pos);
        }
    for(npos = wordinfo->wordhood_list; npos; npos = npos->next){
        for(qpos = npos->neighbour->query_pos_list; qpos;
            qpos = qpos->next){
            /* FIXME: don't use this query_info if single query */
            UGAM_Matcher_WordInfo_seed(qpos->query_info, target_info,
                                       qpos->query_pos, target_pos);
            }
        }
    return;
    }

/* FIXME: not called */
#if 0
static void UGAM_Matcher_WordInfo_destroy(
            UGAM_Matcher *matcher, UGAM_Matcher_WordInfo *wordinfo){
    /* FIXME: empty lists here */
    RecycleBin_recycle(matcher->recycle_wordinfo, wordinfo);
    return;
    }
#endif /* 0 */

static UGAM_Matcher_WordInfo *UGAM_Matcher_WordInfo_merge(
        UGAM_Matcher *matcher,
        UGAM_Matcher_WordInfo *wordinfo_a,
        UGAM_Matcher_WordInfo *wordinfo_b){
    /* Assume wordinfo_b is always empty during merging */
    g_assert(!wordinfo_b->query_pos_list);
    g_assert(!wordinfo_b->wordhood_list);
    RecycleBin_recycle(matcher->recycle_wordinfo, wordinfo_b);
    return wordinfo_a;
    }

/**/

static gpointer UGAM_Matcher_FSM_merge_func(gpointer a, gpointer b,
                                            gpointer user_data){
    register UGAM_Matcher_WordInfo *wordinfo_a = a,
                                   *wordinfo_b = b;
    register UGAM *ugam = user_data;
    g_assert(wordinfo_a);
    g_assert(wordinfo_b);
    g_assert(ugam);
    g_assert(ugam->curr_matcher);
    return UGAM_Matcher_WordInfo_merge(ugam->curr_matcher,
                                       wordinfo_a, wordinfo_b);
    }

static gpointer UGAM_Matcher_FSM_combine_func(gpointer a, gpointer b,
                                              gpointer user_data){
    g_error("UGAM implementation assumes words of the same length");
    return NULL;
    }

static gboolean UGAM_Matcher_decide_fsm_type(gchar *force_fsm){
    if(!g_strcasecmp(force_fsm, "none")){
        /* FIXME: improve
         *     - check if VFSM is possible within memory limit
         *     - look at query sizes ...
         */
        return TRUE;
        }
    if(!g_strcasecmp(force_fsm, "normal"))
        return TRUE;
    if(!g_strcasecmp(force_fsm, "compact"))
        return FALSE;
    g_error("Unknown fsm type [%s]", force_fsm);
    return TRUE;
    }

static UGAM_Matcher *UGAM_Matcher_create(UGAM *ugam){
    register UGAM_Matcher *matcher = g_new(UGAM_Matcher, 1);
    register gchar *alphabet;
    matcher->is_prepared = FALSE;
    matcher->query_info_list = g_ptr_array_new();
    if(ugam->comparison_param->target_param->is_translated)
        alphabet = ugam->param->query_param->alphabet;
    else
        alphabet = ugam->param->target_param->alphabet;
    if(ugam->verbosity > 1)
        g_message("Using comparison alphabet: [%s][%d]",
                  alphabet, strlen(alphabet));
    if(UGAM_Matcher_decide_fsm_type(ugam->uas->force_fsm)){
        matcher->fsm = FSM_create(alphabet,
                              UGAM_Matcher_FSM_merge_func,
                              UGAM_Matcher_FSM_combine_func, ugam);
        matcher->vfsm = NULL;
        matcher->vfsm_leaf = NULL;
        /* Add filter for unmasked target sequences */
        FSM_add_insertion_filter(matcher->fsm,
                Alphabet_get_filter_by_type(
                ugam->comparison_param->comparison_alphabet,
                ugam->comparison_param->cas->softmasked_query
                    ?Alphabet_Filter_Type_MASKED
                    :Alphabet_Filter_Type_UNMASKED));
        FSM_add_traversal_filter(matcher->fsm,
                Alphabet_get_filter_by_type(
                ugam->comparison_param->comparison_alphabet,
                ugam->comparison_param->cas->softmasked_target
                    ?Alphabet_Filter_Type_MASKED
                    :Alphabet_Filter_Type_UNMASKED));
    } else {
        matcher->vfsm = VFSM_create(alphabet, ugam->seed_len);
        if(!matcher->vfsm)
            g_error("Could not create vfsm for alphabet[%s] depth[%d]",
                    alphabet, ugam->seed_len);
        g_message("use [%" VFSM_PRINT_FORMAT "d] for fsm leaf",
                  matcher->vfsm->lrw);
        matcher->vfsm_leaf = g_new0(UGAM_Matcher_WordInfo*,
                                    matcher->vfsm->lrw);
        g_assert(matcher->vfsm_leaf);
        matcher->fsm = NULL;
        }
    matcher->recycle_wordinfo = RecycleBin_create(
            "UGAM_Matcher_WordInfo", sizeof(UGAM_Matcher_WordInfo), 64);
    matcher->recycle_qpos = RecycleBin_create(
            "UGAM_Matcher_QPos", sizeof(UGAM_Matcher_QPos), 64);
    matcher->recycle_npos = RecycleBin_create(
            "UGAM_Matcher_NPos", sizeof(UGAM_Matcher_NPos), 64);
    matcher->comparison_count = 0;
    matcher->query_expect = 0;
    matcher->total_query_length = 0;
    return matcher;
    }

static void UGAM_Matcher_destroy(UGAM_Matcher *matcher){
    register gint i;
    register UGAM_Matcher_QueryInfo *query_info;
    if(matcher->recycle_wordinfo)
        RecycleBin_destroy(matcher->recycle_wordinfo);
    if(matcher->recycle_qpos)
        RecycleBin_destroy(matcher->recycle_qpos);
    if(matcher->recycle_npos)
        RecycleBin_destroy(matcher->recycle_npos);
    if(matcher->fsm)
        FSM_destroy(matcher->fsm);
    if(matcher->vfsm)
        VFSM_destroy(matcher->vfsm);
    if(matcher->vfsm_leaf)
        g_free(matcher->vfsm_leaf);
    for(i = 0; i < matcher->query_info_list->len; i++){
        query_info = matcher->query_info_list->pdata[i];
        g_assert(query_info);
        UGAM_Matcher_QueryInfo_destroy(query_info);
        }
    g_ptr_array_free(matcher->query_info_list, TRUE);
    g_free(matcher);
    return;
    }

static gsize UGAM_Matcher_memory_usage(UGAM_Matcher *matcher){
    register gsize fsm_memory;
    g_assert(matcher);
    if(matcher->fsm)
         fsm_memory = FSM_memory_usage(matcher->fsm);
    else
         fsm_memory = sizeof(VFSM)
                    + (sizeof(UGAM_Matcher_WordInfo)
                       * matcher->vfsm->lrw);
    return sizeof(UGAM_Matcher)
         + (matcher->query_info_list->len
           *(sizeof(gpointer)+sizeof(UGAM_Matcher_QueryInfo)))
         + fsm_memory
         + RecycleBin_memory_usage(matcher->recycle_wordinfo)
         + RecycleBin_memory_usage(matcher->recycle_qpos)
         + RecycleBin_memory_usage(matcher->recycle_npos);
    }
/* Assume no MatchInfo, as used before comparison */

static void UGAM_Matcher_prepare(UGAM_Matcher *matcher){
    g_assert(!matcher->is_prepared);
    /* FIXME: optimisation: convert every WordInfo into compact form ?
     */
    if(matcher->fsm)
        FSM_compile(matcher->fsm);
    matcher->is_prepared = TRUE;
    return;
    }

typedef struct {
                     UGAM *ugam;
         UGAM_StrandParam *param;
    UGAM_Matcher_WordInfo *curr_wordinfo;
} UGAM_Matcher_TraverseData;

static gboolean UGAM_Matcher_WordHood_traverse(gchar *word,
                    gint score, gpointer user_data){
    register UGAM_Matcher_WordInfo *wordinfo;
    register UGAM_Matcher_TraverseData *umtd = user_data;
    register UGAM_Matcher *matcher = umtd->ugam->curr_matcher;
    register VFSM_Int state = 0, leaf;
    g_assert(matcher);
    if(matcher->fsm){ /* Add to FSM */
        wordinfo = UGAM_Matcher_WordInfo_create(matcher);
        wordinfo = FSM_add(matcher->fsm, word, umtd->param->wordlen,
                           wordinfo);
    } else { /* Add to VFSM */
        g_assert(matcher->vfsm);
        g_assert(matcher->vfsm_leaf);
        state = VFSM_word2state(matcher->vfsm, word);
        leaf = VFSM_state2leaf(matcher->vfsm, state);
        wordinfo = matcher->vfsm_leaf[leaf];
        if(!wordinfo){
            wordinfo = UGAM_Matcher_WordInfo_create(matcher);
            matcher->vfsm_leaf[leaf] = wordinfo;
            }
        }
    g_assert(wordinfo);
    UGAM_Matcher_WordInfo_add_npos(umtd->ugam->curr_matcher,
                                   wordinfo, umtd->curr_wordinfo);
    if(wordinfo->query_pos_list)
        UGAM_Matcher_WordInfo_add_npos(umtd->ugam->curr_matcher,
                umtd->curr_wordinfo, wordinfo);
    return FALSE;
    }

static void UGAM_Matcher_QueryInfo_add_query_fsm(UGAM *ugam,
            UGAM_Matcher_QueryInfo *query_info,
            gchar *seq, gint len, gint frame,
            UGAM_StrandParam *param){
    register gint i, pos, orig_pos, valid_count = 0, wj_ctr = 0;
    register UGAM_Matcher_WordInfo *wordinfo;
    UGAM_Matcher_TraverseData umtd;
    g_assert(query_info);
    g_assert(seq);
    g_assert(len >= 0);
    g_assert(ugam);
    g_assert(ugam->curr_matcher);
    umtd.ugam = ugam;
    umtd.param = param;
    for(i = 0; i < len; i++){
        if(param->index[(guchar)seq[i]] != '-'){
            valid_count++;
            if(valid_count >= param->wordlen){
                if(wj_ctr--)
                    continue;
                wj_ctr = ugam->uas->word_jump - 1;
                /* FIXME: apply sat_th here */
                wordinfo = UGAM_Matcher_WordInfo_create(
                                                    ugam->curr_matcher);
                pos = i - param->wordlen + 1;
                g_assert(wordinfo);
                if(frame)
                    orig_pos = (pos * 3) + frame - 1;
                else
                    orig_pos = pos;
                UGAM_Matcher_WordInfo_add_qpos(ugam, ugam->curr_matcher,
                             wordinfo, query_info, orig_pos);
                if(wordinfo->query_pos_list /* not blocked */
                && (!wordinfo->query_pos_list->next)){ /* 1st qpos */
                    umtd.curr_wordinfo = wordinfo;
                    WordHood_traverse(ugam->word_hood,
                               seq+pos, param->wordlen,
                               ugam->param->word_threshold,
                               ugam->uas->use_wordhood_dropoff,
                               UGAM_Matcher_WordHood_traverse, &umtd);
                    }
                }
        } else {
            valid_count = 0;
            }
        }
    return;
    }

/* FIXME: merge two versions ... */
static void UGAM_Matcher_QueryInfo_add_query_vfsm(UGAM *ugam,
            UGAM_Matcher_QueryInfo *query_info,
            gchar *seq, gint len, gint frame,
            UGAM_StrandParam *param){
    register gint i, ch, wj_ctr = 0, pos, orig_pos;
    register VFSM *vfsm = ugam->curr_matcher->vfsm;
    register VFSM_Int state = 0, leaf;
    register UGAM_Matcher_WordInfo *wordinfo;
    UGAM_Matcher_TraverseData umtd;
    umtd.ugam = ugam;
    umtd.param = param;
    g_assert(vfsm);
    for(i = 0; i < len; i++){
        ch = toupper(seq[i]); /* FIXME: filter properly */
        if(!vfsm->index[ch]){
            state = 0;
            continue;
            }
        /* FIXME: use POW2 versions where applicable */
        state = VFSM_change_state_M(vfsm, state, ch);
        if(VFSM_state_is_leaf(vfsm, state)){
            if(wj_ctr--)
                continue;
            wj_ctr = ugam->uas->word_jump - 1;
            /* FIXME: apply sat_th here */
            leaf = VFSM_state2leaf(vfsm, state);
            wordinfo = ugam->curr_matcher->vfsm_leaf[leaf];
            if(!wordinfo)
                wordinfo = UGAM_Matcher_WordInfo_create(
                                                ugam->curr_matcher);
            pos = i - param->wordlen + 1;
            if(frame)
                orig_pos = (pos * 3) + frame - 1;
            else
                orig_pos = pos;
            UGAM_Matcher_WordInfo_add_qpos(ugam, ugam->curr_matcher,
                    wordinfo, query_info, orig_pos);
            /* FIXME: need to implement proper VFSM wordhood */
            if(wordinfo->query_pos_list /* not blocked */
            && (!wordinfo->query_pos_list->next)){ /* 1st qpos */
                umtd.curr_wordinfo = wordinfo;
                WordHood_traverse(ugam->word_hood,
                              seq+pos, param->wordlen,
                              ugam->param->word_threshold,
                              ugam->uas->use_wordhood_dropoff,
                              UGAM_Matcher_WordHood_traverse, &umtd);
                }
            }
        }
    return;
    }

static void UGAM_Matcher_QueryInfo_add_query(UGAM *ugam,
            UGAM_Matcher_QueryInfo *query_info,
            gchar *seq, gint len, gint frame,
            UGAM_StrandParam *param){
    if(ugam->saturate_threshold){
        ugam->curr_matcher->total_query_length += len;
        ugam->curr_matcher->query_expect = UGAM_Matcher_get_expect(
                ugam, ugam->curr_matcher->total_query_length,
                ugam->param->query_param);
        }
    if(ugam->curr_matcher->fsm){
        UGAM_Matcher_QueryInfo_add_query_fsm(ugam, query_info,
                                             seq, len, frame, param);
    } else {
        UGAM_Matcher_QueryInfo_add_query_vfsm(ugam, query_info,
                                             seq, len, frame, param);
        }
    return;
    }
/* WordInfo building algorithm:
 * ---------------------------
 * for each qy_word
 *     if qy_word not in FSM
 *        add empty posinfo to FSM for qy_word
 *     add to qpos_list
 *     if this is first in qpos_list
 *         for each nh_word
 *             if nh_word not in FSM
 *                 add empty posinfo to FSM for nh_word
 *             add curr_word to nb_list of the nh_word
 *             if nh_word->qpos_list is not empty
 *                 add nh_word to qy_word->nb_list
 *
 * Block query list when exceeds saturate_threshold
 */

static void UGAM_Matcher_add_query(UGAM *ugam, Sequence *query){
    register UGAM_Matcher_QueryInfo *query_info
        = UGAM_Matcher_QueryInfo_create(ugam, query);
    register gint i, aa_len;
    register gchar *aa_seq;
    /**/
    g_assert(ugam->curr_matcher);
    g_ptr_array_add(ugam->curr_matcher->query_info_list, query_info);
    if(ugam->comparison_param->query_param->is_translated){
        aa_seq = g_new(gchar, (query->len/3)+2);
        g_assert(ugam->comparison_param->translate);
        for(i = 0; i < 3; i++){
            aa_len = Translate_sequence(
                         ugam->comparison_param->translate,
                         query->seq, query->len,
                         i+1, aa_seq,
               ugam->comparison_param->query_param->alphabet->masked);
            UGAM_Matcher_QueryInfo_add_query(ugam,
                              query_info, aa_seq, aa_len, i+1,
                              ugam->param->target_param);
            }
        g_free(aa_seq);
    } else {
        UGAM_Matcher_QueryInfo_add_query(ugam,
                              query_info, query->seq, query->len, 0,
                              ugam->param->query_param);
        }
    return;
    }

static void UGAM_Matcher_FSM_traverse_func(guint seq_pos,
                                           gpointer node_data,
                                           gpointer user_data){
    register UGAM_Matcher_WordInfo *wordinfo = node_data;
    register UGAM_Matcher_TargetInfo *target_info = user_data;
    register gint tpos;
    if(target_info->curr_frame)
        tpos = (seq_pos * 3) + target_info->curr_frame - 1;
    else
        tpos = seq_pos;
    UGAM_Matcher_WordInfo_traverse(wordinfo, target_info,
                               tpos - target_info->ugam->tpos_modifier);
    return;
    }

static void UGAM_Matcher_VFSM_traverse(UGAM_Matcher *matcher,
                            gchar *seq,
                            UGAM_Matcher_TargetInfo *target_info){
    register gint i, ch;
    register VFSM_Int state = 0, leaf;
    register VFSM *vfsm = matcher->vfsm;
    register UGAM_Matcher_WordInfo *wordinfo;
    g_assert(vfsm);
    for(i = 0; seq[i]; i++){
        ch = toupper(seq[i]); /* FIXME: filter properly */
        if(!vfsm->index[ch]){
            state = 0;
            continue;
            }
        /* FIXME: use POW2 versions where applicable */
        state = VFSM_change_state_M(vfsm, state, ch);
        if(VFSM_state_is_leaf(vfsm, state)){
            leaf = VFSM_state2leaf(vfsm, state);
            wordinfo = matcher->vfsm_leaf[leaf];
            if(wordinfo)
                UGAM_Matcher_FSM_traverse_func(i, wordinfo,
                                               target_info);
            }
        }
    return;
    }

static void UGAM_Matcher_compare_target(UGAM *ugam,
                                        UGAM_Matcher *matcher,
                                        Sequence *target){
    register gint i, aa_len;
    register gchar *aa_seq;
    UGAM_Matcher_TargetInfo target_info;
    register UGAM_Matcher_QueryInfo *query_info;
    target_info.ugam = ugam;
    target_info.target = Sequence_share(target);
    target_info.queryinfo_list = g_ptr_array_new();
    if(ugam->comparison_param->target_param->is_translated){
        aa_seq = g_new(gchar, (target->len/3)+2);
        for(i = 0; i < 3; i++){
            target_info.curr_frame = i+1;
            aa_len = Translate_sequence(
                        ugam->comparison_param->translate,
                        target->seq, target->len, i+1, aa_seq,
                ugam->comparison_param->target_param->alphabet->masked);
            if(matcher->fsm)
                FSM_traverse(matcher->fsm, aa_seq,
                         UGAM_Matcher_FSM_traverse_func, &target_info);
            else
                UGAM_Matcher_VFSM_traverse(matcher, aa_seq,
                                           &target_info);
            }
        g_free(aa_seq);
    } else {
        target_info.curr_frame = 0;
        if(matcher->fsm)
            FSM_traverse(matcher->fsm, target->seq,
                     UGAM_Matcher_FSM_traverse_func, &target_info);
        else
            UGAM_Matcher_VFSM_traverse(matcher, target->seq,
                                       &target_info);
        }
    Sequence_destroy(target_info.target);
    for(i = 0; i < target_info.queryinfo_list->len; i++){
        query_info = target_info.queryinfo_list->pdata[i];
        g_assert(query_info->curr_comparison);
        if(!HSPset_is_empty(query_info->curr_comparison->hspset) > 0){
            HSPset_finalise(query_info->curr_comparison->hspset);
            ugam->report_func(query_info->curr_comparison,
                              ugam->user_data);
            }
        Comparison_destroy(query_info->curr_comparison);
        query_info->curr_comparison = NULL;
        }
    g_ptr_array_free(target_info.queryinfo_list, TRUE);
    return;
    }

/**/

UGAM_ArgumentSet *UGAM_ArgumentSet_create(Argument *arg){
    register ArgumentSet *as;
    static UGAM_ArgumentSet uas;
    if(arg){
        as = ArgumentSet_create("Ungapped Alignment Options");
        ArgumentSet_add_option(as, 'M', "fsmmemory", "Mb",
            "Memory limit for FSM scanning", "64",
            Argument_parse_int, &uas.fsm_memory_limit);
        ArgumentSet_add_option(as, '\0', "forcefsm", "fsm type",
            "Force FSM type ( normal | compact )", "none",
            Argument_parse_string, &uas.force_fsm);
        ArgumentSet_add_option(as, 0, "useworddropoff", NULL,
            "Use word neighbourhood dropoff", "TRUE",
            Argument_parse_boolean, &uas.use_wordhood_dropoff);
        ArgumentSet_add_option(as, 0, "wordjump", NULL,
            "Jump betwen wordneighbourhood words", "1",
            Argument_parse_int, &uas.word_jump);
        /**/
        ArgumentSet_add_option(as, 0, "dnawordthreshold", "score",
            "Threshold score for word neighbourhood generation", "5",
            Argument_parse_int, &uas.dna_word_threshold);
        ArgumentSet_add_option(as, 0, "proteinwordthreshold", "score",
            "Threshold score for word neighbourhood generation", "3",
            Argument_parse_int, &uas.protein_word_threshold);
        /**/
        Argument_absorb_ArgumentSet(arg, as);
        }
    return &uas;
    }

/**/

gboolean UGAM_query_present(UGAM *ugam, FastaDB_Seq *query){
    g_assert(ugam);
    g_assert(query);
    g_assert(query->seq->alphabet->type
          == ugam->comparison_param->query_param->alphabet->type);
    /**/
    g_assert(ugam->curr_matcher);
    UGAM_Matcher_add_query(ugam, query->seq);
    /**/
    if(ugam->verbosity)
        g_message("Memory usage: [%d] [Mb:%d]",
                  UGAM_Matcher_memory_usage(ugam->curr_matcher),
                  UGAM_Matcher_memory_usage(ugam->curr_matcher)>>20);
    if(UGAM_Matcher_memory_usage(ugam->curr_matcher)
     > (ugam->uas->fsm_memory_limit << 20))
        return TRUE;
    return FALSE;
    }

/**/

void UGAM_target_present(UGAM *ugam, FastaDB_Seq *target){
    g_assert(ugam);
    g_assert(target);
    if(ugam->verbosity)
        g_message("Finding matches with target [%s]", target->seq->id);
    UGAM_Matcher_compare_target(ugam, ugam->curr_matcher, target->seq);
    if(ugam->verbosity)
        g_message("Done finding matches with target [%s]",
                target->seq->id);
    return;
    }

/**/

void UGAM_query_prepare(UGAM *ugam){
    ugam->curr_matcher = UGAM_Matcher_create(ugam);
    return;
    }

void UGAM_target_prepare(UGAM *ugam){
    g_assert(ugam->curr_matcher);
    UGAM_Matcher_prepare(ugam->curr_matcher);
    return;
    }

void UGAM_clear(UGAM *ugam){
    g_assert(ugam->curr_matcher);
    UGAM_Matcher_destroy(ugam->curr_matcher);
    ugam->curr_matcher = NULL;
    return;
    }

/**/

static UGAM_StrandParam *UGAM_StrandParam_create(gchar *alphabet,
                             guint wordlen, gboolean is_softmasked){
    register UGAM_StrandParam *usp = g_new(UGAM_StrandParam, 1);
    register gint i;
    usp->alphabet = alphabet;
    usp->alphabet_size = strlen(alphabet);
    usp->wordlen = wordlen;
    for(i = 0; i < ALPHABET_SIZE; i++)
        usp->index[i] = '-';
    for(i = 0; i < usp->alphabet_size; i++){
        usp->index[(guchar)usp->alphabet[i]] = usp->alphabet[i];
        if(!is_softmasked)
            usp->index[tolower(usp->alphabet[i])] = usp->alphabet[i];
        }
    return usp;
    }

static void UGAM_StrandParam_destroy(UGAM_StrandParam *usp){
    g_free(usp);
    return;
    }

/**/

#define UGAM_GET_AnalysisType(qt, tt) (((qt) << 8) | (tt))

static UGAM_Param *UGAM_Param_create(Comparison_Param *comparison_param,
                             UGAM_ArgumentSet *uas,
                             gint dna_wordlen, gint protein_wordlen,
                             gint verbosity){
    register Alphabet_Type
        query_type = comparison_param->query_param->alphabet->type,
        target_type = comparison_param->target_param->alphabet->type;
    register gboolean translate_both
              = (comparison_param->query_param->is_translated
              && comparison_param->target_param->is_translated);
    register UGAM_Param *up = g_new(UGAM_Param, 1);
    static gchar *dna_alphabet = "ACGT",
                 *protein_alphabet = "ARNDCQEGHILKMFPSTWYV";
    if(translate_both){
        g_assert(query_type == Alphabet_Type_DNA);
        g_assert(target_type == Alphabet_Type_DNA);
        if(verbosity)
            g_message("Translated DNA:DNA comparison");
        up->query_param = UGAM_StrandParam_create(
                          protein_alphabet, protein_wordlen,
                        comparison_param->cas->softmasked_query);
        up->target_param = UGAM_StrandParam_create(
                            protein_alphabet, protein_wordlen,
                            comparison_param->cas->softmasked_target);
        up->word_threshold = uas->protein_word_threshold;
    } else {
        switch(UGAM_GET_AnalysisType(query_type, target_type)){
            case UGAM_GET_AnalysisType(Alphabet_Type_DNA,
                                       Alphabet_Type_DNA):
                if(verbosity)
                    g_message("Simple DNA:DNA comparison");
                up->query_param = UGAM_StrandParam_create(
                                   dna_alphabet, dna_wordlen,
                        comparison_param->cas->softmasked_query);
                up->target_param = UGAM_StrandParam_create(
                                   dna_alphabet, dna_wordlen,
                            comparison_param->cas->softmasked_target);
                up->word_threshold = uas->dna_word_threshold;
                break;
            case UGAM_GET_AnalysisType(Alphabet_Type_PROTEIN,
                                       Alphabet_Type_DNA):
                if(verbosity)
                    g_message("PROTEIN query vs Translated DNA target");
                up->query_param = UGAM_StrandParam_create(
                                   protein_alphabet, protein_wordlen,
                        comparison_param->cas->softmasked_query);
                up->target_param = UGAM_StrandParam_create(
                                   dna_alphabet, protein_wordlen*3,
                            comparison_param->cas->softmasked_target);
                up->word_threshold = uas->protein_word_threshold;
                break;
            case UGAM_GET_AnalysisType(Alphabet_Type_DNA,
                                       Alphabet_Type_PROTEIN):
                if(verbosity)
                    g_message(
                            "Translating DNA query vs PROTEIN target");
                up->query_param = UGAM_StrandParam_create(
                                   dna_alphabet, protein_wordlen*3,
                        comparison_param->cas->softmasked_query);
                up->target_param = UGAM_StrandParam_create(
                                   protein_alphabet, protein_wordlen,
                            comparison_param->cas->softmasked_target);
                up->word_threshold = uas->protein_word_threshold;
                break;
            case UGAM_GET_AnalysisType(Alphabet_Type_PROTEIN,
                                       Alphabet_Type_PROTEIN):
                if(verbosity)
                    g_message("Simple PROTEIN:PROTEIN comparison");
                up->query_param = UGAM_StrandParam_create(
                                   protein_alphabet, protein_wordlen,
                        comparison_param->cas->softmasked_query);
                up->target_param = UGAM_StrandParam_create(
                                   protein_alphabet, protein_wordlen,
                            comparison_param->cas->softmasked_target);
                up->word_threshold = uas->protein_word_threshold;
                break;
            default:
                g_error("Unknown query type [%d] with target type [%d]",
                         query_type, target_type);
                break;
            }
        }
    return up;
    }

static void UGAM_Param_destroy(UGAM_Param *up){
    UGAM_StrandParam_destroy(up->query_param);
    UGAM_StrandParam_destroy(up->target_param);
    g_free(up);
    return;
    }

/**/

/* FIXME: if(scan_query)
 *     - fastapipe had query and target swapped
 *     - UGAM gets scan_query argument
 *            and must swap comparsion_param
 *            but use orig for HSP set
 */

UGAM *UGAM_create(Comparison_Param *comparison_param,
                  UGAM_ReportFunc report_func, gpointer user_data,
                  gint dna_wordlen, gint protein_wordlen,
                  gint saturate_threshold, gboolean scan_query,
                  gint verbosity){
    register UGAM *ugam = g_new(UGAM, 1);
    g_assert(comparison_param);
    g_assert(report_func);
    if(scan_query)
        ugam->comparison_param
                = Comparison_Param_swap(comparison_param);
    else
        ugam->comparison_param
                = Comparison_Param_share(comparison_param);
    if(verbosity)
        g_message("Creating UGAM with query[%s] target[%s]",
            Alphabet_Type_get_name(ugam->comparison_param->query_param
                                   ->alphabet->type),
            Alphabet_Type_get_name(ugam->comparison_param->target_param
                                   ->alphabet->type));
    ugam->uas = UGAM_ArgumentSet_create(NULL);
    ugam->param = UGAM_Param_create(ugam->comparison_param, ugam->uas,
                                    dna_wordlen, protein_wordlen,
                                    verbosity);
    ugam->report_func = report_func;
    ugam->scan_query = scan_query;
    ugam->user_data = user_data;
    ugam->verbosity = verbosity;
    ugam->saturate_threshold = saturate_threshold;
    if(ugam->comparison_param->target_param->is_translated){
        ugam->word_hood = WordHood_create(
                              ugam->param->query_param->alphabet,
                              ugam->param->query_param->alphabet,
                              ugam->comparison_param->submat,
               ugam->comparison_param->cas->softmasked_query);
    } else {
        ugam->word_hood = WordHood_create(
                              ugam->param->target_param->alphabet,
                              ugam->param->target_param->alphabet,
                              ugam->comparison_param->submat,
               ugam->comparison_param->cas->softmasked_query);
        }
    ugam->curr_matcher = NULL;
    /* Set the target position modifier */
    if(ugam->comparison_param->target_param->is_translated){
        if(ugam->comparison_param->query_param->is_translated){
            /* Q,T */
            ugam->tpos_modifier =
                + (ugam->param->query_param->wordlen * 3) - 3;
            ugam->seed_len = ugam->param->query_param->wordlen;
        } else { /* -,T */
            ugam->tpos_modifier = ugam->param->target_param->wordlen
                                - 3;
            ugam->seed_len = MIN(ugam->param->query_param->wordlen,
                                 ugam->param->target_param->wordlen);
            }
        /* -,? */
    } else {
        ugam->tpos_modifier = ugam->param->target_param->wordlen - 1;
        ugam->seed_len = MIN(ugam->param->query_param->wordlen,
                             ugam->param->target_param->wordlen);
        }
    return ugam;
    }

void UGAM_destroy(UGAM *ugam){
    Comparison_Param_destroy(ugam->comparison_param);
    UGAM_Param_destroy(ugam->param);
    if(ugam->curr_matcher)
        UGAM_Matcher_destroy(ugam->curr_matcher);
    WordHood_destroy(ugam->word_hood);
    g_free(ugam);
    return;
    }

