/****************************************************************\
*                                                                *
*  Codon Usage Module                                            *
*                                                                *
*  Guy St.C. Slater..   mailto:guy@ebi.ac.uk                     *
*  Copyright (C) 2000-2005.  All Rights Reserved.                *
*                                                                *
*  This source code is distributed under the terms of the        *
*  GNU Lesser General Public License. See the file COPYING       *
*  or http://www.fsf.org/copyleft/lesser.html for details        *
*                                                                *
*  If you use this code, please keep this notice intact.         *
*                                                                *
\****************************************************************/

#include <math.h> /* For log() */

#ifndef ALPHABETSIZE
#include <limits.h>
#define ALPHABETSIZE sizeof(gchar) << CHAR_BIT
#endif /* ALPHABETSIZE */

#include "codonusage.h"
#include "vfsm.h"

static gint codon_usage_human_composition [4] =
    /* A C G T */ { 7347209, 7460888, 7562638, 6251713 };

static gint codon_usage_human[4][4][4] = {
{
  /* AAA AAC AAG AAT */ { 234181, 181317, 303300, 161921 },
  /* ACA ACC ACG ACT */ { 142653, 176216,  57615, 124086 },
  /* AGA AGC AGG AGT */ { 115112, 188190, 112356, 117586 },
  /* ATA ATC ATG ATT */ {  71847, 195129, 209437, 151184 }
},
{ /* CAA CAC CAG CAT */ { 119315, 144696, 330754, 102499 },
  /* CCA CCC CCG CCT */ { 165666, 193691,  69083, 170770 },
  /* CGA CGC CGG CGT */ {  59486, 101057, 112142,  43776 },
  /* CTA CTC CTG CTT */ {  69270, 185667, 379418, 125766 }
},
{ /* GAA GAC GAG GAT */ { 282562, 240998, 382475, 209886 },
  /* GCA GCC GCG GCT */ { 155155, 269612,  72209, 175872 },
  /* GGA GGC GGG GGT */ { 157139, 214449, 157816, 102466 },
  /* GTA GTC GTG GTT */ {  67196, 135296, 264842, 104156 }
},
{ /* TAA TAC TAG TAT */ {   5542, 142285,   4524, 114024 },
  /* TCA TCC TCG TCT */ { 117025, 167291,  43389, 142843 },
  /* TGA TGC TGG TGT */ {   9654, 118116, 120335,  98968 },
  /* TTA TTC TTG TTT */ {  72997, 190646, 122166, 163696 }
}
    };

/* ATG: TAA TAG TGA */

static void CodonUsage_expand_word_count_recur(VFSM *vfsm,
               VFSM_Int state, gint count, gint id,
               gint *vfsm_count, gint *vfsm_mailbox){
    register gchar *word = g_new(gchar, vfsm->depth+1);
    register gint i;
    register gchar c;
    register VFSM_Int curr;
    VFSM_state2word(vfsm, state, word);
    vfsm_mailbox[VFSM_state2leaf(vfsm, state)] = id;
    vfsm_count[VFSM_state2leaf(vfsm, state)] += count;
    for(i = 0; i < vfsm->depth; i++){
        if(word[i] != 'N'){
            c = word[i];
            word[i] = 'N';
            curr = VFSM_word2state(vfsm, word);
            if(vfsm_mailbox[VFSM_state2leaf(vfsm, curr)] != id)
                CodonUsage_expand_word_count_recur(vfsm,
                        curr, count, id, vfsm_count, vfsm_mailbox);
            word[i] = c;
            }
        }
    g_free(word);
    return;
    }

static gdouble CodonUsage_chance_word_probability(gchar *word,
                 gint wordlen, gdouble *compprob){
    register gint i;
    register gdouble cp = compprob[(guchar)*word];
    g_assert(word);
    g_assert(wordlen);
    for(i = 1; i < wordlen; i++)
        cp *= compprob[(guchar)word[i]];
    return cp;
    }

static void CodonUsage_print_scores(VFSM *vfsm, gint *vfsm_count){
    register gint i;
    register gchar *word = g_new0(gchar, vfsm->depth+1);
    register gdouble rawprob, chanceprob, odds_score;
    register gint counttotal = vfsm_count[vfsm->lrw-1];
    register double logbase = log(2.0);
    gdouble compprob[ALPHABETSIZE] = {0.0};
    compprob['N'] += compprob['A'] = codon_usage_human_composition[0];
    compprob['N'] += compprob['C'] = codon_usage_human_composition[1];
    compprob['N'] += compprob['G'] = codon_usage_human_composition[2];
    compprob['N'] += compprob['T'] = codon_usage_human_composition[3];
    compprob['A'] /= compprob['N'];
    compprob['C'] /= compprob['N'];
    compprob['G'] /= compprob['N'];
    compprob['T'] /= compprob['N'];
    compprob['N'] /= compprob['N'];
    for(i = 0; i < vfsm->lrw; i++){
        VFSM_state2word(vfsm, VFSM_leaf2state(vfsm, i), word);
        if(vfsm_count[i]){
            rawprob = ((gdouble)vfsm_count[i])/((gdouble)counttotal);
        } else { /* Avoid -inf probabilities where count == 0 */
            rawprob = 0.48/((gdouble)counttotal);
            }
        chanceprob = CodonUsage_chance_word_probability(
                word, vfsm->depth, compprob);
        odds_score= rawprob/chanceprob;
        g_message("word [%s] count [%d] rp[%f] cp[%f]"
                  " odd_score[%f] los[%f]",
                word, vfsm_count[i], rawprob, chanceprob,
                odds_score,
                log(odds_score)/logbase);
        }
    g_free(word);
    return;
    }

static void CodonUsage_expand_word_count(void){
    register gchar *alphabet = "ACGTN";
    register VFSM *vfsm = VFSM_create(alphabet, 3);
    register gint i, j, k, id = 0;
    register gchar *word = g_new0(gchar, 4);
    register gint *vfsm_count = g_new0(gint, vfsm->lrw),
                  *vfsm_mailbox  = g_new0(gint, vfsm->lrw);
    for(i = 0; i < 4; i++){
        word[0] = alphabet[i];
        for(j = 0; j < 4; j++){
            word[1] = alphabet[j];
            for(k = 0; k < 4; k++){
                word[2] = alphabet[k];
                CodonUsage_expand_word_count_recur(vfsm,
                        VFSM_word2state(vfsm, word),
                        codon_usage_human[i][j][k], ++id,
                        vfsm_count, vfsm_mailbox);
                }
            }
        }
    CodonUsage_print_scores(vfsm, vfsm_count);
    VFSM_destroy(vfsm);
    g_free(vfsm_count);
    g_free(vfsm_mailbox);
    return;
    }

CodonUsage *CodonUsage_create(void){
    register CodonUsage *cu = g_new(CodonUsage, 1);
    CodonUsage_expand_word_count();
    return cu;
    }

void CodonUsage_destroy(CodonUsage *cu){
    g_free(cu);
    return;
    }

