/****************************************************************\
*                                                                *
*  Efficient double-linked list routines.                        *
*                                                                *
*  Guy St.C. Slater..   mailto:guy@ebi.ac.uk                     *
*  Copyright (C) 2000-2008.  All Rights Reserved.                *
*                                                                *
*  This source code is distributed under the terms of the        *
*  GNU General Public License, version 3. See the file COPYING   *
*  or http://www.gnu.org/licenses/gpl.txt for details            *
*                                                                *
*  If you use this code, please keep this notice intact.         *
*                                                                *
\****************************************************************/

#include "dlist.h"

/**/

DListSet *DListSet_create(void){
    register DListSet *dlist_set = g_new(DListSet, 1);
    dlist_set->list_recycle = RecycleBin_create("DList",
                                                sizeof(DList), 256);
    dlist_set->node_recycle = RecycleBin_create("DListNode",
                                           sizeof(DListNode), 1024);
    return dlist_set;
    }

void DListSet_destroy(DListSet *dlist_set){
    g_assert(dlist_set);
    RecycleBin_destroy(dlist_set->list_recycle);
    RecycleBin_destroy(dlist_set->node_recycle);
    g_free(dlist_set);
    return;
    }

gsize DListSet_memory_usage(DListSet *dlist_set){
    return sizeof(DListSet)
         + RecycleBin_memory_usage(dlist_set->list_recycle)
         + RecycleBin_memory_usage(dlist_set->node_recycle);
    }

/**/

static DListNode *DListNode_alloc(DListSet *dlist_set){
    register DListNode *node
                         = RecycleBin_alloc(dlist_set->node_recycle);
    node->next = NULL;
    node->prev = NULL;
    node->data = NULL;
    return node;
    }

static void DListNode_free(DListSet *dlist_set, DListNode *node){
    RecycleBin_recycle(dlist_set->node_recycle, node);
    return;
    }

/**/

DList *DList_create(DListSet *dlist_set){
    register DList *dlist;
    g_assert(dlist_set);
    dlist = RecycleBin_alloc(dlist_set->list_recycle);
    dlist->head = DListNode_alloc(dlist_set);
    dlist->tail = DListNode_alloc(dlist_set);
    dlist->set = dlist_set;
    dlist->head->next = dlist->tail;
    dlist->tail->prev = dlist->head;
    return dlist;
    }

void DList_empty(DList *dlist){
    g_assert(dlist);
    while(!DList_isempty(dlist))
        DList_pop_head(dlist);
    return;
    }

void DList_destroy(DList *dlist){
    g_assert(dlist);
    DList_empty(dlist);
    RecycleBin_recycle(dlist->set->list_recycle, dlist);
    return;
    }

void DList_empty_with_data(DList *dlist,
                           DList_DestroyFunc destroy_func,
                           gpointer user_data){
    register DListNode *node;
    g_assert(dlist);
    g_assert(destroy_func);
    while(!DList_isempty(dlist)){
        node = DList_pop_head(dlist);
        destroy_func(node, user_data);
        }
    return;
    }

void DList_destroy_with_data(DList *dlist,
                             DList_DestroyFunc destroy_func,
                             gpointer user_data){
    g_assert(dlist);
    g_assert(destroy_func);
    DList_empty_with_data(dlist, destroy_func, user_data);
    DList_destroy(dlist);
    return;
    }

DListNode *DList_stack(DList *dlist, gpointer data){
    register DListNode *node = DListNode_alloc(dlist->set);
    g_assert(dlist);
    node->data = data;
    node->prev = dlist->head;
    node->next = dlist->head->next;
    dlist->head->next = node->next->prev = node;
    return node;
    }

DListNode *DList_queue(DList *dlist, gpointer data){
    register DListNode *node = DListNode_alloc(dlist->set);
    g_assert(dlist);
    node->data = data;
    node->next = dlist->tail;
    node->prev = dlist->tail->prev;
    dlist->tail->prev = node->prev->next = node;
    return node;
    }

gpointer DList_pop_head(DList *dlist){
    register DListNode *node = dlist->head->next;
    register gpointer data = node->data;
    g_assert(dlist);
    dlist->head->next = node->next;
    dlist->head->next->prev = dlist->head;
    DListNode_free(dlist->set, node);
    return data;
    }

gpointer DList_pop_tail(DList *dlist){
    register DListNode *node = dlist->tail->prev;
    register gpointer data = node->data;
    g_assert(dlist);
    dlist->tail->prev = node->prev;
    dlist->tail->prev->next = dlist->tail;
    DListNode_free(dlist->set, node);
    return data;
    }

DList *DList_join(DList *left, DList *right){
    register DListNode *head, *tail;
    g_assert(left);
    g_assert(right);
    g_assert(left->set == right->set);
    if(DList_isempty(right))
        return left;
    if(DList_isempty(left)){
        head = left->head;
        tail = left->tail;
        left->head = right->head;
        left->tail = right->tail;
        right->head = head;
        right->tail = tail;
        return left;
        }
    g_assert(!DList_isempty(left));
    g_assert(!DList_isempty(right));
    /**/
    left->tail->prev->next = right->head->next;
    right->head->next->prev = left->tail->prev;
    left->tail->prev = right->tail->prev;
    left->tail->prev->next = left->tail;
    /**/
    right->head->next = right->tail;
    right->tail->prev = right->head;
    return left;
    }

gpointer DList_remove(DList *dlist, DListNode *node){
    register gpointer data;
    g_assert(dlist);
    g_assert(node);
    node->prev->next = node->next;
    node->next->prev = node->prev;
    data = node->data;
    DListNode_free(dlist->set, node);
    return data;
    }

DListNode *DList_insert_after(DList *dlist, DListNode *prev_node,
                              gpointer data){
    register DListNode *node = DListNode_alloc(dlist->set);
    g_assert(dlist);
    if(!prev_node)
        prev_node = dlist->head->next;
    node->data = data;
    node->prev = prev_node;
    node->next = prev_node->next;
    prev_node->next->prev = node;
    prev_node->next = node;
    return node;
    }

DListNode *DList_insert_before(DList *dlist, DListNode *next_node,
                        gpointer data){
    register DListNode *node = DListNode_alloc(dlist->set);
    g_assert(dlist);
    if(!next_node)
        next_node = dlist->tail->prev;
    node->data = data;
    node->next = next_node;
    node->prev = next_node->prev;
    next_node->prev->next = node;
    next_node->prev = node;
    return node;
    }

void DList_traverse(DList *dlist, DList_TraverseFunc traverse_func,
                    gpointer user_data){
    register DListNode *node;
    g_assert(dlist);
    g_assert(traverse_func);
    DList_for_each(dlist, node){
        if(traverse_func(node->data, user_data))
            break;
        }
    return;
    }

void DList_raise(DList *dlist, DListNode *node){
    node->prev->next = node->next;
    node->next->prev = node->prev;
    node->prev = dlist->head;
    node->next = dlist->head->next;
    dlist->head->next = node->next->prev = node;
    return;
    }

void DList_lower(DList *dlist, DListNode *node){
    node->prev->next = node->next;
    node->next->prev = node->prev;
    node->next = dlist->tail;
    node->prev = dlist->tail->prev;
    dlist->tail->prev = node->prev->next = node;
    return;
    }

/**/

