/****************************************************************\
*                                                                *
*  Efficient double-linked list routines.                        *
*                                                                *
*  Guy St.C. Slater..   mailto:guy@ebi.ac.uk                     *
*  Copyright (C) 2000-2008.  All Rights Reserved.                *
*                                                                *
*  This source code is distributed under the terms of the        *
*  GNU General Public License, version 3. See the file COPYING   *
*  or http://www.gnu.org/licenses/gpl.txt for details            *
*                                                                *
*  If you use this code, please keep this notice intact.         *
*                                                                *
\****************************************************************/

/* Note: DList is used instead of the glib GList object,
 * as it provides:
 *   o Support for insertion mid-list not easy with glib-1
 *   o All constant time operations (including adding and joining)
 *     ( except DList_traverse() and DList_destroy_with_data() )
 *   o Ease of destroying a set of lists at once
 *   o Ability to add to either end of list.
 *   o DList does not change on append (has a container)
 */

#ifndef INCLUDED_DLIST_H
#define INCLUDED_DLIST_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <glib.h>
#include "recyclebin.h"

typedef struct DListNode {
    struct DListNode *prev;
    struct DListNode *next;
            gpointer  data;
} DListNode;

typedef struct {
    RecycleBin *list_recycle;
    RecycleBin *node_recycle;
} DListSet;

typedef struct {
    DListNode *head;
    DListNode *tail;
     DListSet *set;
} DList;
/* FIXME: optimisation : remove dlist->set to save space */

#define DList_isempty(dlist) ((dlist)->head->next == (dlist)->tail)
#define DList_first(dlist) \
        (DList_isempty(dlist)?NULL:(dlist)->head->next->data)
#define DList_last(dlist) \
        (DList_isempty(dlist)?NULL:(dlist)->tail->prev->data)
#define DList_for_each(dlist, node)   \
    for((node) = (dlist)->head->next; \
        (node) != (dlist)->tail;      \
        (node) = (node)->next)

typedef void (*DList_DestroyFunc)(gpointer data, gpointer user_data);
typedef gboolean (*DList_TraverseFunc)(gpointer data,
                                       gpointer user_data);
/* DList_TraverseFunc should return TRUE to stop the traversal
 */

DListSet *DListSet_create(void);
    void  DListSet_destroy(DListSet *dlist_set);
   gsize  DListSet_memory_usage(DListSet *dlist_set);
/* DListSet also destroy will all DLists in this set */

DList *DList_create(DListSet *dlist_set);
 void  DList_empty(DList *dlist);
 void  DList_destroy(DList *dlist);
 void  DList_destroy_with_data(DList *dlist,
                               DList_DestroyFunc destroy_func,
                               gpointer user_data);
 void  DList_empty_with_data(DList *dlist,
                             DList_DestroyFunc destroy_func,
                             gpointer user_data);

DListNode *DList_stack(DList *dlist, gpointer data);
DListNode *DList_queue(DList *dlist, gpointer data);
 gpointer  DList_pop_head(DList *dlist);
 gpointer  DList_pop_tail(DList *dlist);

  DList *DList_join(DList *left, DList *right);
/* DList_stack() adds to start of list
 * DList_queue() adds to end of list
 * DList_pop_head() removes from the start of the list
 * DList_pop_tail() removes from the end of the list
 * DList_join() appends right onto end of left.
 *              right is emptied but not destroyed.
 */

gpointer DList_remove(DList *dlist, DListNode *node);
DListNode *DList_insert_after(DList *dlist, DListNode *prev_node,
                        gpointer data);
DListNode *DList_insert_before(DList *dlist, DListNode *next_node,
                         gpointer data);
/* DList_remove() removes node from dlist and returns node data
 * DList_insert_after() inserts data into dlist after prev_node
 *                      (or at the beggining if prev_node is NULL)
 * DList_insert_before() inserts data into dlist before next_node
 *                       (or at the start if next_node is NULL)
 */

void DList_traverse(DList *dlist, DList_TraverseFunc traverse_func,
                    gpointer user_data);

void DList_raise(DList *dlist, DListNode *node);
void DList_lower(DList *dlist, DListNode *node);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* INCLUDED_DLIST_H */

